/*
* extrusn.cpp
*
* $Id: extrusn.cpp,v 1.6 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* This module creates Extrusion objects.
*/

#include <math.h>
#include <string.h>

#include "pov.h"
#include "povout.h"

using namespace std;

#define PROBLEM(WHAT) do { \
	cerr << "Yikes! Problem with Extrusion: " \
	<< WHAT << '\n'; \
	return; \
} while (0)

inline int is_length_zero(F_TYPE *v)   // allow for small digital error
{
	return ((v[0]*v[0]+v[1]*v[1]+v[2]*v[2]) < (1.0e-12));
}

inline int is_same_point(F_TYPE *v1, F_TYPE *v2, int n)
{
	if (n==3)
		return (fabs((v1[0]-v2[0])) < (1.0e-6) &&
		fabs((v1[1]-v2[1])) < (1.0e-6) &&
		fabs((v1[2]-v2[2])) < (1.0e-6));
	else
		return (fabs((v1[0]-v2[0])) < (1.0e-6) &&
		fabs((v1[1]-v2[1])) < (1.0e-6));
}

inline void normalize(F_TYPE *v)  // make unit length
{
	F_TYPE length=sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	for (int i=0; i<3; i++) v[i]/=length;
}

class SCP;

class Triangle {
public:
	SCP *s1, *s2, *s3;
	int i1, i2, i3;
	F_TYPE n1[3], n2[3], n3[3];
};

class SCP {
public:
	F_TYPE x[3], y[3], z[3];   // unit vectors for local coordinate system
	F_TYPE *points;	     // crossSection points in real 3-d coordinates
	DLList<Triangle> **shared;
	int n_csps;
	SCP(int n) {
		n_csps=n;
		int i;
		for (i=0; i<3; i++) x[i]=y[i]=z[i]=0.0;
		points=new F_TYPE[3*n];
		if (!points) {
			cerr << "Ran out of memory\n";
            abort();
		}
		shared=new DLList<Triangle>*[n];
		if (!shared) {
			cerr << "Ran out of memory\n";
            abort();
		}
		for (i=0; i<n; i++) {
			shared[i]=new DLList<Triangle>;
			if (!shared[i]) {
				cerr << "Ran out of memory\n";
                abort();
			}
		}
	}
	~SCP() {
		delete points;
		for (int i=0; i<n_csps; i++) delete shared[i];
		delete shared;
	}
};

void do_Extrusion(SFNode* node)
{
	F_TYPE *csps=(FIND(node,crossSection,MFVec2f))->value; // crossSection points
	int n_csps=(FIND(node,crossSection,MFVec2f))->n_items;
	int closed_crossSection=is_same_point(&(csps[0]),&(csps[2*(n_csps-1)]),2);
	
	F_TYPE *orients=(FIND(node,orientation,MFRotation))->value;
	int n_orients=(FIND(node,orientation,MFRotation))->n_items;
	
	F_TYPE *scales=(FIND(node,scale,MFVec2f))->value;
	int n_scales=(FIND(node,scale,MFVec2f))->n_items;
	
	F_TYPE *sps=(FIND(node,spine,MFVec3f))->value; // spine points
	int n_sps=(FIND(node,spine,MFVec3f))->n_items;
	if (n_sps<2) PROBLEM("need at least 2 spine points");
	int closed_spine=is_same_point(&(sps[0]),&(sps[3*(n_sps-1)]),3);
	
	if (n_scales==0 ||
		(n_scales>1 && n_scales<n_sps))
		PROBLEM("not enough scale vectors");
	if (n_orients==0 ||
		(n_orients>1 && n_orients<n_sps))
		PROBLEM("not enough orientation vectors");
	
	int ccw=FIND(node,ccw,SFBool)->value;
	int beginCap=FIND(node,beginCap,SFBool)->value;
	int endCap=FIND(node,endCap,SFBool)->value;
	int convex=FIND(node,convex,SFBool)->value;
	if (verbose && !convex && (beginCap || endCap))
		cerr << "Warning: non-convex polygons not yet supported;\n"
		<< "  Extrusion caps will not be generated.\n";
	if (!convex) beginCap=endCap=FALSE;
	int solid=FIND(node,solid,SFBool)->value;
	if (verbose && solid)
		cerr <<
		"Warning: all polygons being used despite Extrusion being solid.\n";
	F_TYPE creaseAngle=FIND(node,creaseAngle,SFFloat)->value;
	//  creaseAngle=0.0;  // turns off smoothing
	
	// allocate memory for required objects
	
	int i, j;
	SCP **scps=new SCP*[n_sps];
	if (!scps) MEMOUT;
	for (i=0; i<n_sps; i++) {
		scps[i]=new SCP(n_csps);
		if (!scps[i]) MEMOUT;
	}
	
	int n_tris=2*(n_sps-1)*(n_csps-1);
	int n_cap_tris= (closed_crossSection) ? n_csps-3 : n_csps-2;
	if (beginCap) n_tris+=n_cap_tris;
	if (endCap) n_tris+=n_cap_tris;
	Triangle **tris=new Triangle*[n_tris];
	if (!tris) MEMOUT;
	for (i=0; i<n_tris; i++) {
		tris[i]=new Triangle();
		if (!tris[i]) MEMOUT;
	}
	
	//
	// The algoritm use here to construct the Extrusion is taken directly
	// out of the VRML specification...
	//
	
	// first compute the SCP Y axis unit vectors at each spine point
	
	for (i=1; i<n_sps-1; i++) {
		scps[i]->y[0]=sps[3*(i+1)]-sps[3*(i-1)];
		scps[i]->y[1]=sps[3*(i+1)+1]-sps[3*(i-1)+1];
		scps[i]->y[2]=sps[3*(i+1)+2]-sps[3*(i-1)+2];
		if (is_length_zero(scps[i]->y))
			PROBLEM("spine points i+1 and i-1 can't be the same point");
		else
			normalize(scps[i]->y);
	}
	if (closed_spine) {
		scps[0]->y[0]=sps[3]-sps[3*(n_sps-2)];
		scps[0]->y[1]=sps[4]-sps[3*(n_sps-2)+1];
		scps[0]->y[2]=sps[5]-sps[3*(n_sps-2)+2];
		if (is_length_zero(scps[0]->y))
			PROBLEM("spine points 1 and n-2 can't be the same point");
		else
			normalize(scps[0]->y);
		for (i=0; i<3; i++) scps[n_sps-1]->y[i]=scps[0]->y[i];
	} else {
		scps[0]->y[0]=sps[3]-sps[0];
		scps[0]->y[1]=sps[4]-sps[1];
		scps[0]->y[2]=sps[5]-sps[2];
		if (is_length_zero(scps[0]->y))
			PROBLEM("spine points 1 and 0 can't be the same point");
		else
			normalize(scps[0]->y);
		scps[n_sps-1]->y[0]=sps[3*(n_sps-1)]-sps[3*(n_sps-2)];
		scps[n_sps-1]->y[1]=sps[3*(n_sps-1)+1]-sps[3*(n_sps-2)+1];
		scps[n_sps-1]->y[2]=sps[3*(n_sps-1)+2]-sps[3*(n_sps-2)+2];
		if (is_length_zero(scps[n_sps-1]->y))
			PROBLEM("spine points n-1 and n-2 can't be the same point");
		else
			normalize(scps[n_sps-1]->y);
	}
	// then find the Z axes
	
	F_TYPE f[3], b[3];
	for (i=1; i<n_sps-1; i++) {
		for (j=0; j<3; j++) {
			f[j]=sps[3*(i+1)+j]-sps[3*i+j];
			b[j]=sps[3*(i-1)+j]-sps[3*i+j];
		}
		scps[i]->z[0]=f[1]*b[2]-f[2]*b[1];
		scps[i]->z[1]=f[2]*b[0]-f[0]*b[2];
		scps[i]->z[2]=f[0]*b[1]-f[1]*b[0];
		if (!is_length_zero(scps[i]->z))
			normalize(scps[i]->z);
	}
	if (closed_spine) {
		for (j=0; j<3; j++) {
			f[j]=sps[3+j]-sps[j];
			b[j]=sps[3*(n_sps-2)+j]-sps[j];
		}
		scps[0]->z[0]=f[1]*b[2]-f[2]*b[1];
		scps[0]->z[1]=f[2]*b[0]-f[0]*b[2];
		scps[0]->z[2]=f[0]*b[1]-f[1]*b[0];
		if (!is_length_zero(scps[i]->z)) normalize(scps[i]->z);
		for(j=0; j<3; j++) scps[n_sps-1]->z[j]=scps[0]->z[j];
	} else {
		for(j=0; j<3; j++) scps[0]->z[j]=scps[1]->z[j];
		for(j=0; j<3; j++) scps[n_sps-1]->z[j]=scps[n_sps-2]->z[j];
	}
	
	// deal with special cases for Z axes
	
	if (is_length_zero(scps[0]->z)) {
		for (i=1; i<n_sps; i++) {
			if (!is_length_zero(scps[i]->z)) {
				for (j=0; j<3; j++) scps[0]->z[j]=scps[i]->z[j];
				break;
			}
		}
	}
	
	for (i=1; i<n_sps; i++) {
		if (is_length_zero(scps[i]->z)) {
			for (j=0; j<3; j++) scps[i]->z[j]=scps[i-1]->z[j];
		}
	}
	
	//
	// if whole spine is collinear, find Rotation that takes positive Y axis into
	// the vector from spine[0] to spine[n-1], then apply that rotation to
	// the y=0 plane to get the SCP axes.
	//
	
	Matrix m;
	j=0;
	for (i=0; i<n_sps; i++) {
		if (!is_length_zero(scps[i]->z)) j++;
	}
	if (j<n_sps) {
		if (j>0) PROBLEM("internal error: some, but not all, SCP Z-axes defined");
		Vector x, y, z;
		x.v[0]=1;
		y.v[1]=1;
		z.v[2]=1;
		F_TYPE s[3];
        SFRotation *r = new SFRotation();
		for (i=0; i<3; i++) s[i]=sps[3*(n_sps-1)+i]-sps[i];
        r->value[0]=s[2];
        r->value[1]=0.0;      // simplified version of r = y cross s
        r->value[2]=-s[0];
        if (is_length_zero(r->value)) {
			if (s[1]<0.0) { // if spine is along -Y axis,
				x.v[0]=-1;    // rotate SCP 180 deg about Z axis
				y.v[1]=-1;    // (note VRML spec. doesn't define what to do here...)
			}
		} else {
			normalize(s);
            r->value[3]=acos(s[1]);
			set_rotation_matrix(&m,r);
			apply_transformation(&m,&x);
			apply_transformation(&m,&y);
			apply_transformation(&m,&z);
		}
		for (i=0; i<n_sps; i++) {
			for (j=0; j<3; j++) {
				scps[i]->x[j]=x.v[j];
				scps[i]->y[j]=y.v[j];
				scps[i]->z[j]=z.v[j];
			}
		}
	} else {
		
		// determine X axes (= Y cross Z)
		
		for (i=1; i<n_sps; i++) {
			if ((((scps[i]->z[0])*(scps[i-1]->z[0]))+
				((scps[i]->z[1])*(scps[i-1]->z[1]))+
				((scps[i]->z[2])*(scps[i-1]->z[2])))<0.0)
				for (j=0; j<3; j++) scps[i]->z[j]*=-1.0;
		}
		for (i=0; i<n_sps; i++) {
			scps[i]->x[0]=(scps[i]->y[1])*(scps[i]->z[2])-
				(scps[i]->y[2])*(scps[i]->z[1]);
			scps[i]->x[1]=(scps[i]->y[2])*(scps[i]->z[0])-
				(scps[i]->y[0])*(scps[i]->z[2]);
			scps[i]->x[2]=(scps[i]->y[0])*(scps[i]->z[1])-
				(scps[i]->y[1])*(scps[i]->z[0]);
		}
	}
	
	// now compute location of crossSection points at each spine point
	
	SFRotation *rot = new SFRotation(); // nodes are deleted later
	int k;
	for (i=0; i<n_sps; i++) {     // for each spine point
		for (j=0; j<n_csps; j++) {  // for each cross section point
			
			// first position points with orientation of SCP, and scale
			
			if (n_scales>1) {
				for (k=0; k<3; k++) {
					scps[i]->points[3*j+k]=scps[i]->x[k] * csps[2*j] * scales[2*i];
					scps[i]->points[3*j+k]+=scps[i]->z[k] * csps[2*j+1] * scales[2*i+1];
				}
			} else {
				for (k=0; k<3; k++) {
					scps[i]->points[3*j+k]=scps[i]->x[k] * csps[2*j] * scales[0];
					scps[i]->points[3*j+k]+=scps[i]->z[k] * csps[2*j+1] * scales[1];
				}
			}
			
			// then apply orientation relative to SCP local coordinates
			
			if (j==0) {
				if (n_orients>1) {
					for (k=0; k<3; k++) {
                        rot->value[k]=scps[i]->x[k] * orients[4*i];
                        rot->value[k]+=scps[i]->y[k] * orients[4*i+1];
                        rot->value[k]+=scps[i]->z[k] * orients[4*i+2];
					}
                    rot->value[3]=orients[4*i+3];
				} else {
					for (k=0; k<3; k++) {
                        rot->value[k]=scps[i]->x[k] * orients[0];
                        rot->value[k]+=scps[i]->y[k] * orients[1];
                        rot->value[k]+=scps[i]->z[k] * orients[2];
					}
                    rot->value[3]=orients[3];
				}
				set_rotation_matrix(&m,rot);
			}
			apply_transformation(&m,&(scps[i]->points[3*j]));
			
			// finally, translate by spine coordinate
			
			for (k=0; k<3; k++) scps[i]->points[3*j+k] += sps[3*i+k];
			
		}
	}
	
	//
	// Now that we have all the coordinates of the crossSection points, just link
	// them up to form triangles. Also store lists of triangles to use in
	// normal calculations.
	//
	
	for (i=0; i<n_sps-1; i++) {
		for (j=0; j<n_csps-1; j++) {
			int ind=2*(i*(n_csps-1)+j);
			
			tris[ind]->s1=tris[ind]->s2=scps[i];
			tris[ind]->s3=scps[i+1];
			tris[ind]->i1=tris[ind]->i3=j;
			tris[ind]->i2=j+1;
			
			tris[ind+1]->s1=scps[i];
			tris[ind+1]->s2=tris[ind+1]->s3=scps[i+1];
			tris[ind+1]->i3=j;
			tris[ind+1]->i1=tris[ind+1]->i2=j+1;
			
			if (creaseAngle>0.0) {
				scps[i]->shared[j]->add(tris[ind]);
				scps[i]->shared[j+1]->add(tris[ind]);
				scps[i+1]->shared[j]->add(tris[ind]);
				
				scps[i]->shared[j+1]->add(tris[ind+1]);
				scps[i+1]->shared[j+1]->add(tris[ind+1]);
				scps[i+1]->shared[j]->add(tris[ind+1]);
			}
		}
	}
	
	// put on caps
	
	if (beginCap) {
		j=2*(n_sps-1)*(n_csps-1);
		for (i=0; i<n_cap_tris; i++) {
			tris[j+i]->s1=tris[j+i]->s2=tris[j+i]->s3=scps[0];
			tris[j+i]->i1=0;
			tris[j+i]->i2=i+2;
			tris[j+i]->i3=i+1;
			scps[0]->shared[0]->add(tris[j+1]);
			scps[0]->shared[i+2]->add(tris[j+1]);
			scps[0]->shared[i+1]->add(tris[j+1]);
		}
	}
	if (endCap) {
		j=2*(n_sps-1)*(n_csps-1);
		if (beginCap) j+=n_cap_tris;
		for (i=0; i<n_cap_tris; i++) {
			tris[j+i]->s1=tris[j+i]->s2=tris[j+i]->s3=scps[n_sps-1];
			tris[j+i]->i1=0;
			tris[j+i]->i2=i+1;
			tris[j+i]->i3=i+2;
			scps[n_sps-1]->shared[0]->add(tris[j+1]);
			scps[n_sps-1]->shared[i+1]->add(tris[j+1]);
			scps[n_sps-1]->shared[i+2]->add(tris[j+1]);
		}
	}
	
	//
	// Do normal calculation if necessary. First, calculate all triangle
	// normals. Then, for each vertex of each triangle, the normal at that
	// point is the average of the normals of all triangles that share that
	// vertex and whose normals are within creaseAngle of each other.
	//
	
	if (creaseAngle>0.0) {
		Triangle* tri;
		F_TYPE a[3], b[3];
		F_TYPE *n=NULL;
		F_TYPE cos_creaseAngle=cos(creaseAngle);
		
		// first, calculate all triangle unit normals
		
		for (i=0; i<n_tris; i++) {
			tri=tris[i];
			n=tri->n1;     // for now store triangle's normal in first normal field
			if (ccw) {
				for (j=0; j<3; j++) a[j]=tri->s3->points[3*(tri->i3)+j] -
					tri->s2->points[3*(tri->i2)+j];
				for (j=0; j<3; j++) b[j]=tri->s1->points[3*(tri->i1)+j] -
					tri->s2->points[3*(tri->i2)+j];
			} else {
				for (j=0; j<3; j++) b[j]=tri->s3->points[3*(tri->i3)+j] -
					tri->s2->points[3*(tri->i2)+j];
				for (j=0; j<3; j++) a[j]=tri->s1->points[3*(tri->i1)+j] -
					tri->s2->points[3*(tri->i2)+j];
			}
			n[0]=a[1]*b[2]-a[2]*b[1];
			n[1]=a[2]*b[0]-a[0]*b[2];   // n = a X b
			n[2]=a[0]*b[1]-a[1]*b[0];
            if (!is_length_zero(n)) normalize(n);
		}
		
		// for each vertex, find all triangles that share this vertex
		
		n=new F_TYPE[3*n_tris];	      // buffer for first vertex normal
		if (!n) MEMOUT;
		int k, l, ci, nn;
		F_TYPE *cn=NULL, *on;
		SCP* cs;
		for (i=0; i<n_tris; i++) {			// loop through triangles
			tri=tris[i];
			for (j=0; j<3; j++) {			// loop through vertexes
				if (j==0) { cn=&(n[3*i]); cs=tri->s1; ci=tri->i1; }
				else if (j==1) { cn=tri->n2; cs=tri->s2; ci=tri->i2; }
				else { cn=tri->n3; cs=tri->s3; ci=tri->i3; }
				for (l=0; l<3; l++) cn[l]=tri->n1[l];
				nn=1;
				
				for (k=0; k<(cs->shared[ci]->n_items()); k++) {
					
					//
					// and average values if angle between normals < creaseAngle. Calculate
					// dot product of normals: if dot product (=cos(angle)) > cos(creaseAngle)
					// then angle < creaseAngle. (Saves having to do (slow) cos calculation
					// on every single pair of normals...)
					//
					
					if ((cs->shared[ci]->current())!=tri) {
						on=cs->shared[ci]->current()->n1;
						F_TYPE dotp=tri->n1[0]*on[0]+tri->n1[1]*on[1]+tri->n1[2]*on[2];
						if (dotp>cos_creaseAngle) {
							for (l=0; l<3; l++) cn[l]+=on[l];
							nn++;
						}
					}
					cs->shared[ci]->next();
				}
				for (l=0; l<3; l++) cn[l]/=((F_TYPE)nn);
			}
		}
		for (i=0; i<n_tris; i++) {
			for (j=0; j<3; j++)
				tris[i]->n1[j]=n[3*i+j]; // copy n1 from temp buffer
		}
		delete n;
	}
	
	// Print out the triangle mesh
	
    if (n_tris) {
        out << INDENT << "mesh {\n";
        indent+=ii;
        for (i=0; i<n_tris; i++) {
            if (creaseAngle>0.0) {
                out << INDENT << "smooth_triangle {\n";
                indent+=ii;
                out << INDENT << '<' << tris[i]->s1->points[3*(tris[i]->i1)] << ','
                    << tris[i]->s1->points[3*(tris[i]->i1)+1] << ','
                    << tris[i]->s1->points[3*(tris[i]->i1)+2] << ">, "
                    << '<' << tris[i]->n1[0] << ',' << tris[i]->n1[1]
                    << ',' << tris[i]->n1[2] << ">\n"
				
                    << INDENT << '<' << tris[i]->s2->points[3*(tris[i]->i2)] << ','
                    << tris[i]->s2->points[3*(tris[i]->i2)+1] << ','
                    << tris[i]->s2->points[3*(tris[i]->i2)+2] << ">, "
                    << '<' << tris[i]->n2[0] << ',' << tris[i]->n2[1]
                    << ',' << tris[i]->n2[2] << ">\n"
				
                    << INDENT << '<' << tris[i]->s3->points[3*(tris[i]->i3)] << ','
                    << tris[i]->s3->points[3*(tris[i]->i3)+1] << ','
                    << tris[i]->s3->points[3*(tris[i]->i3)+2] << ">, "
                    << '<' << tris[i]->n3[0] << ',' << tris[i]->n3[1]
                    << ',' << tris[i]->n3[2] << ">\n";
                indent-=ii;
                out << INDENT << "}\n";
            } else {
                out << INDENT << "triangle { "
                    << '<' << tris[i]->s1->points[3*(tris[i]->i1)] << ','
                    << tris[i]->s1->points[3*(tris[i]->i1)+1] << ','
                    << tris[i]->s1->points[3*(tris[i]->i1)+2] << ">, "
                    << '<' << tris[i]->s2->points[3*(tris[i]->i2)] << ','
                    << tris[i]->s2->points[3*(tris[i]->i2)+1] << ','
                    << tris[i]->s2->points[3*(tris[i]->i2)+2] << ">, "
                    << '<' << tris[i]->s3->points[3*(tris[i]->i3)] << ','
                    << tris[i]->s3->points[3*(tris[i]->i3)+1] << ','
                    << tris[i]->s3->points[3*(tris[i]->i3)+2] << "> }\n";
            }
        }
        out << INDENT << current_transformation << '\n';
        if (finish[0]) out << INDENT << finish << '\n';
        indent-=ii;
        out << INDENT << "}\n";
    }
	
	// clean up.
	
	for (i=0; i<n_sps; i++) delete scps[i];
	delete scps;
	for (i=0; i<n_tris; i++) delete tris[i];
	delete tris;
	
	return;
}
