/*
* fieldtyp.h
*
* $Id: fieldtyp.h,v 1.5 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*            http://www.chemicalgraphics.com/paul
*
* This contains the C++ classes defined for each VRML97 type. I have tried
* as much as possible to use variable names analogous to those defined in
* the VRML97 specification grammar.
*
* should put in desctructors, eventually, just for completeness...
*/

#ifndef __vrml2pov_fieldtyp_h_
#define __vrml2pov_fieldtyp_h_

#include <string.h>
#include <iostream>

#include "global.h"
#include "list.h"
#include "node.h"

class fieldType {
private:
    static DLList<fieldType> allocatedNodes; // held in node.cpp
public:
	char fieldTypeName;  // for type checking
    virtual ~fieldType() {}
    // for memory management -- all nodes allocated are kept in this list
    // and should only be deallocated by FreeAllocatedNodes(). This is because
    // pointers to nodes are duplicated freely around the parse tree, so it's
    // otherwise hard to ensure they're deleted only once.
    fieldType() { allocatedNodes.add(this); }
    static void FreeAllocatedNodes() { allocatedNodes.empty_and_delete_all(); }
};

enum { EVENT=-1, ERROR=0, SFBOOL, SFCOLOR, MFCOLOR, SFFLOAT, MFFLOAT, SFINT32,
	SFIMAGE, MFINT32, SFNODE, MFNODE, SFROTATION, MFROTATION, SFSTRING,
	MFSTRING, SFTIME, MFTIME, SFVEC2F, MFVEC2F, SFVEC3F, MFVEC3F };

#define NOMEM do { std::cerr << "Ran out of memory\n"; abort(); } while (0)

// For now this is just an empty field of type EVENT...
class Event : public fieldType {
public:
	Event() { fieldTypeName=EVENT; }
};

class SFBool : public fieldType {
public:
	I_TYPE value;
	SFBool(I_TYPE v =FALSE) { fieldTypeName=SFBOOL; value=v; }
};

class SFColor : public fieldType {
public:
	F_TYPE value[3];
	SFColor(F_TYPE r =0.0, F_TYPE g =0.0, F_TYPE b=0.0) {
		fieldTypeName=SFCOLOR;
		value[0]=r;
		value[1]=g;
		value[2]=b;
	}
};

class MFColor : public fieldType {
public:
	F_TYPE *value;
	int n_items;
	MFColor(int n =0) {
		fieldTypeName=MFCOLOR;
		if (n) {
			value=new F_TYPE[3*n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFColor() { if (value) delete value; }
};

class SFFloat : public fieldType {
public:
	F_TYPE value;
	SFFloat(F_TYPE v =0.0) { fieldTypeName=SFFLOAT; value=v; }
};

class MFFloat : public fieldType {
public:
	F_TYPE* value;
	int n_items;
	MFFloat(int n =0) {
		fieldTypeName=MFFLOAT;
		if (n) {
			value=new F_TYPE[n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFFloat() { if (value) delete value; }
};

//
// Pixels are stored as integer values; so this only works if the #bytes
// in an integer is <= # byte-components of each pixel...
//
#define SFIMAGE_INT_TYPE int
class SFImage : public fieldType {
public:
	int width, height, n_components;
	SFIMAGE_INT_TYPE *pixel_data;
	SFImage(int w =0, int h=0, int n =0) {
		fieldTypeName=SFIMAGE;
		if (w*h) {
			pixel_data=new SFIMAGE_INT_TYPE[w*h];
			if (!pixel_data) NOMEM;
		} else
			pixel_data=NULL;
		width=w;
		height=h;
		n_components=n;
	}
    ~SFImage() { if (pixel_data) delete pixel_data; }
};

//
// I suppose to be purist these should be ensured to be 32-bit data types, but
// I don't really care that much...
//

class SFInt32 : public fieldType {
public:
	I_TYPE value;
	SFInt32(I_TYPE v =0) { fieldTypeName=SFINT32; value=v; }
};

class MFInt32 : public fieldType {
public:
	I_TYPE* value;
	int n_items;
	MFInt32(int n =0) {
		fieldTypeName=MFINT32;
		if (n) {
			value=new I_TYPE[n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFInt32() { if (value) delete value; }
};

class nodeVariable; // defined in node.h
class nodeISarg;

class SFNode : public fieldType {
public:
	DLList<nodeVariable> nodeBody;
	DLList<nodeISarg> *nodeBodyArgs;
	char* nodeTypeId;
	SFNode(const char* n =NULL) {
		fieldTypeName=SFNODE;
		if (n) {
			nodeTypeId=new char[strlen(n)+1];
			if (!nodeTypeId) NOMEM;
			strcpy(nodeTypeId,n);
		} else
			nodeTypeId=NULL;
		nodeBodyArgs=NULL;
    }
    ~SFNode() {
        nodeBody.empty_and_delete_all();
        if (nodeTypeId) delete nodeTypeId;
        if (nodeBodyArgs) {
            nodeBodyArgs->empty_and_delete_all();
            delete nodeBodyArgs;
        }
    }
};

class MFNode : public fieldType {
public:
	SFNode **value;
	int n_items;
	MFNode(int n =0) {
		fieldTypeName=MFNODE;
		if (n) {
			value=new SFNode*[n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFNode() {
        if (value) delete value;
    }
};

class SFRotation : public fieldType {
public:
	F_TYPE value[4];
	SFRotation(F_TYPE x =0.0, F_TYPE y =0.0, F_TYPE z =1.0, F_TYPE r =0.0) {
		fieldTypeName=SFROTATION;
		value[0]=x;
		value[1]=y;
		value[2]=z;
		value[3]=r;
	}
};

class MFRotation : public fieldType {
public:
	F_TYPE *value;
	int n_items;
	MFRotation(int n =0) {
		fieldTypeName=MFROTATION;
		if (n) {
			value=new F_TYPE[4*n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFRotation() { if (value) delete value; }
};

class SFString : public fieldType {
public:
	char* value;
	SFString(const char* v =NULL) {
		fieldTypeName=SFSTRING;
		if (v) {
			value=new char[strlen(v)+1];
			if (value)
				strcpy(value,v);
			else
				NOMEM;
		} else
			value=NULL;
	}
    ~SFString() { if (value) delete value; }
};

class MFString : public fieldType {
public:
	char **value;
	int n_items;
	MFString(int n =0) {
		fieldTypeName=MFSTRING;
		if (n) {
			value=new char*[n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFString() {
        if (value) {
            for (int i=0; i<n_items; i++)
                if (value[i]) delete value[i];
            delete value;
        }
    }
};

class SFTime : public fieldType {
public:
	F_TYPE value;
	SFTime(F_TYPE v =0.0) { fieldTypeName=SFTIME; value=v; }
};

class MFTime : public fieldType {
public:
	F_TYPE* value;
	int n_items;
	MFTime(int n =0) {
		fieldTypeName=MFTIME;
		if (n) {
			value=new F_TYPE[n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFTime() { if (value) delete value; }
};

class SFVec2f : public fieldType {
public:
	F_TYPE value[2];
	SFVec2f(F_TYPE v1 =0.0, F_TYPE v2 =0.0) {
		fieldTypeName=SFVEC2F;
		value[0]=v1;
		value[1]=v2;
	}
};

class MFVec2f : public fieldType {
public:
	F_TYPE *value;
	int n_items;
	MFVec2f(int n =0) {
		fieldTypeName=MFVEC2F;
		if (n) {
			value=new F_TYPE[2*n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFVec2f() { if (value) delete value; }
};

class SFVec3f : public fieldType {
public:
	F_TYPE value[3];
	SFVec3f(F_TYPE v1 =0.0, F_TYPE v2 =0.0, F_TYPE v3 =0.0) {
		fieldTypeName=SFVEC3F;
		value[0]=v1;
		value[1]=v2;
		value[2]=v3;
	}
};

class MFVec3f : public fieldType {
public:
	F_TYPE *value;
	int n_items;
	MFVec3f(int n =0) {
		fieldTypeName=MFVEC3F;
		if (n) {
			value=new F_TYPE[3*n];
			if (!value) NOMEM;
		} else
			value=NULL;
		n_items=n;
	}
    ~MFVec3f() { if (value) delete value; }
};

unsigned char get_type(const char*);  // implemented in readtype.cc
fieldType* field_copy(fieldType* old);

#endif // __vrml2pov_fieldtyp_h_
