/*
* ifs2mesh.cpp
*
* $Id: ifs2mesh.cpp,v 1.7 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* This routine converts an IndexedFaceSet into a POV mesh of (usually
* smooth) triangles. IndexedFaceSet polygons are reduced to triangles,
* since pov can only do shading smoothing on triangles. Also, that way we
* can store the whole set in a POV mesh{} object.
*/

#include <math.h>
#include <string.h>

#include "pov.h"
#include "povout.h"

using namespace std;

#define PROBLEM(WHAT) do { \
	cerr << "Yikes! Problem with IndexedFaceSet: " \
	<< WHAT << '\n'; \
	return; \
} while (0)

inline int is_length_zero(F_TYPE *v)   // allow for small digital error
{
	return ((v[0]*v[0]+v[1]*v[1]+v[2]*v[2]) < (1.0e-12));
}

inline int is_same_point(F_TYPE *v1, F_TYPE *v2, int n)
{
	if (n==3)
		return (fabs((v1[0]-v2[0])) < (1.0e-6) &&
		fabs((v1[1]-v2[1])) < (1.0e-6) &&
		fabs((v1[2]-v2[2])) < (1.0e-6));
	else
		return (fabs((v1[0]-v2[0])) < (1.0e-6) &&
		fabs((v1[1]-v2[1])) < (1.0e-6));
}

inline void normalize(F_TYPE *v)  // make unit length
{
	F_TYPE length=sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	for (int i=0; i<3; i++) v[i]/=length;
}

class Triangle {
public:
	int i1, i2, i3;
    F_TYPE *n1, *n2, *n3;   // vertex normals
    F_TYPE *fn;             // face normal
	F_TYPE color[3];
	unsigned char culled;
    Triangle() { culled=FALSE; n1=n2=n3=fn=NULL; }
    ~Triangle() { if (fn) delete fn; }
};
static DLList<F_TYPE> allocatedNormals;

void do_IndexedFaceSet(SFNode* node)
{
	SFNode* tmp;

	// first gather all the info in the node

	F_TYPE* coords=NULL;
	int n_coords=0;
	tmp=FIND(node,coord,SFNode);
	if (tmp) {
		coords=(FIND(tmp,point,MFVec3f))->value;
		n_coords=(FIND(tmp,point,MFVec3f))->n_items;
	}
	if (n_coords<3 || !coords) {
		cerr << "Error in IndexedFaceSet - can't have < 3 coordinates!\n";
		return;
	}

	F_TYPE* colors=NULL;
	int n_colors=0;
	tmp=FIND(node,color,SFNode);
	if (tmp) {
		colors=(FIND(tmp,color,MFColor))->value;
		n_colors=(FIND(tmp,color,MFColor))->n_items;
	}

	F_TYPE* normals=NULL;
	int n_normals=0;
	tmp=FIND(node,normal,SFNode);
	if (tmp) {
		normals=(FIND(tmp,vector,MFVec3f))->value;
		n_normals=(FIND(tmp,vector,MFVec3f))->n_items;
	}

	tmp=FIND(node,texCoord,SFNode);
	if (verbose && tmp)
		cerr << "Warning: textured IndexFaceSet not yet implemented...\n";

	int ccw=FIND(node,ccw,SFBool)->value;
	int colorPerVertex=FIND(node,colorPerVertex,SFBool)->value;
	if (verbose && colors && colorPerVertex)
		cerr
		<< "Warning: POV doesn't support coloring polygons by vertex,\n"
		<< "  so polygon color will be assigned average vertex color.\n";
	int convex=FIND(node,convex,SFBool)->value;
    if (verbose && !convex)
		cerr <<
        "Warning: non-convex polygons not supported; results may be unexpected.\n";
	F_TYPE creaseAngle=FIND(node,creaseAngle,SFFloat)->value;
	int normalPerVertex=FIND(node,normalPerVertex,SFBool)->value;
	int solid=FIND(node,solid,SFBool)->value;
	if (verbose && solid && !do_backface_culling)
		cerr <<
		"Warning: all polygons being used despite IndexedFaceSet being solid.\n";

	I_TYPE* coordIndex=(FIND(node,coordIndex,MFInt32))->value;
	int n_coordIndexes=(FIND(node,coordIndex,MFInt32))->n_items;

	I_TYPE* colorIndex=(FIND(node,colorIndex,MFInt32))->value;
	int n_colorIndexes=(FIND(node,colorIndex,MFInt32))->n_items;

	I_TYPE* normalIndex=(FIND(node,normalIndex,MFInt32))->value;
	int n_normalIndexes=(FIND(node,normalIndex,MFInt32))->n_items;

	static DLList<Triangle> triangle_list;
	Triangle* tri;
	int ts=0, tn, face_n=0, i, j;

	//
	// for each vertex, store a list of which triangles share that vertex,
	// for use in normal calculation (if smoothing is to be done).
	//

	int do_smoothing=(!normals && creaseAngle>0.00001);
    int smooth_tri=((normals && normalPerVertex) || do_smoothing);
    
	DLList<Triangle> **shared_vertexes=NULL;
	if (do_smoothing) {
		shared_vertexes=new DLList<Triangle>*[n_coords];
		if (!shared_vertexes) MEMOUT;
		for (i=0; i<n_coords; i++) {
			shared_vertexes[i]=new DLList<Triangle>;
			if (!shared_vertexes[i]) MEMOUT;
		}
	}

	do {			  // loop through polygons in coordIndex
		if ((ts+2)>=n_coordIndexes) PROBLEM("not enough coords for polygon");
		tn=ts+1;

		do {		  // loop through triangles in each polygon
			if (coordIndex[ts]>=n_coords || coordIndex[ts]<0 ||
				coordIndex[tn]>=n_coords || coordIndex[tn]<0 ||
				coordIndex[tn+1]>=n_coords || coordIndex[tn+1]<0)
				PROBLEM("coordIndex out of range of coords");
			tri=new Triangle();
			if (!tri) MEMOUT;
			tri->i1=coordIndex[ts];
			tri->i2=coordIndex[tn];
			tri->i3=coordIndex[tn+1];

			if (smooth_tri) { 	     // collect or calculate normals
				if (do_smoothing) {	     // store triangles for normal calculation
					shared_vertexes[coordIndex[ts]]->add(tri);
					shared_vertexes[coordIndex[tn]]->add(tri);
					shared_vertexes[coordIndex[tn+1]]->add(tri);
				} else {		     // normals already present
					if (n_normalIndexes>0) {
						if (normalIndex[ts]>=n_normals || normalIndex[ts]<0 ||
							normalIndex[tn]>=n_normals || normalIndex[tn]<0 ||
							normalIndex[tn+1]>=n_normals || normalIndex[tn+1]<0)
							PROBLEM("normalIndex out of range of normals");
						tri->n1=&(normals[3*normalIndex[ts]]);
						tri->n2=&(normals[3*normalIndex[tn]]);
						tri->n3=&(normals[3*normalIndex[tn+1]]);
					} else {
						if (coordIndex[ts]>=n_normals || coordIndex[ts]<0 ||
							coordIndex[tn]>=n_normals || coordIndex[tn]<0 ||
							coordIndex[tn+1]>=n_normals || coordIndex[tn+1]<0)
							PROBLEM("coordIndex out of range of normals");
						tri->n1=&(normals[3*coordIndex[ts]]);
						tri->n2=&(normals[3*coordIndex[tn]]);
						tri->n3=&(normals[3*coordIndex[tn+1]]);
					}
				}
			}

			if (colors) {	  // assign colors if present
				if (colorPerVertex) {
					F_TYPE *c1, *c2, *c3;
					if (n_colorIndexes>0) {
						if (colorIndex[ts]>=n_colors || colorIndex[ts]<0 ||
							colorIndex[tn]>=n_colors || colorIndex[tn]<0 ||
							colorIndex[tn+1]>=n_colors || colorIndex[tn+1]<0)
							PROBLEM("colorIndex out of range of colors");
						c1=&(colors[3*colorIndex[ts]]);
						c2=&(colors[3*colorIndex[tn]]);
						c3=&(colors[3*colorIndex[tn+1]]);
					} else {
						if (coordIndex[ts]>=n_colors || coordIndex[ts]<0 ||
							coordIndex[tn]>=n_colors || coordIndex[tn]<0 ||
							coordIndex[tn+1]>=n_colors || coordIndex[tn+1]<0)
							PROBLEM("coordIndex out of range of colors");
						c1=&(colors[3*coordIndex[ts]]);
						c2=&(colors[3*coordIndex[tn]]);
						c3=&(colors[3*coordIndex[tn+1]]);
					}
					for (i=0; i<3; i++) tri->color[i]=(c1[i]+c2[i]+c3[i])/3.0;
				} else {
					if (n_colorIndexes>0) {
						if (face_n>=n_colorIndexes)
							PROBLEM("not enough color indexes for each face");
						if (colorIndex[face_n]>=n_colors || colorIndex[face_n]<0)
							PROBLEM("color index out of range");
						for (i=0; i<3; i++)
							tri->color[i]=colors[3*colorIndex[face_n]+i];
					} else {
						if (face_n>=n_colors) PROBLEM("not enough colors for each face");
						for (i=0; i<3; i++)
							tri->color[i]=colors[3*face_n+i];
					}
				}
			}

			triangle_list.add(tri);
			tn+=1;
		} while ((tn+1)<n_coordIndexes &&
			coordIndex[tn+1]!=-1);
		face_n++;
		ts=tn+2;
	} while (ts<n_coordIndexes);

	//
	// Do normal calculation if necessary. First, calculate all triangle
	// normals. Then, for each vertex of each triangle, the normal at that
	// point is the average of the normals of all triangles that share that
	// vertex and whose normals are within creaseAngle of each other.
	//

	F_TYPE a[3], b[3];
    F_TYPE cos_creaseAngle=cos(creaseAngle);
    if (do_smoothing || do_backface_culling) {

        // first, calculate all triangle face normals

		for (i=0; i<triangle_list.n_items(); i++) {
			tri=triangle_list.current();
            tri->fn=new F_TYPE[3];
            if (!tri->fn) MEMOUT;
			if (ccw) {
				for (j=0; j<3; j++) a[j]=coords[3*(tri->i3)+j]-coords[3*(tri->i2)+j];
				for (j=0; j<3; j++) b[j]=coords[3*(tri->i1)+j]-coords[3*(tri->i2)+j];
			} else {
				for (j=0; j<3; j++) b[j]=coords[3*(tri->i3)+j]-coords[3*(tri->i2)+j];
				for (j=0; j<3; j++) a[j]=coords[3*(tri->i1)+j]-coords[3*(tri->i2)+j];
			}
            tri->fn[0]=a[1]*b[2]-a[2]*b[1];
            tri->fn[1]=a[2]*b[0]-a[0]*b[2];   // n = a X b
            tri->fn[2]=a[0]*b[1]-a[1]*b[0];
            if (!is_length_zero(tri->fn)) normalize(tri->fn);

			// if backface culling is on, check to see if triangle should be displayed

			if (do_backface_culling) {
				static Matrix m;
				static F_TYPE vp[3];
				if (i==0) {
					invert_into(current_transformation,&m); // xform culling viewpoint ...
					for (j=0; j<3; j++) vp[j]=culling_viewpoint[j];
					apply_transformation(&m,vp);            // ... into local coordinates
				}
				F_TYPE dotp=0.0;    // compute dot product of normal with vector to vp
                for (j=0; j<3; j++) dotp+=(tri->fn[j])*(vp[j]-coords[3*(tri->i1)+j]);
				if (dotp<=0.0)      // facing away -> culled
					tri->culled=TRUE;
			}

			triangle_list.next();
		}
	}

	// for each vertex, find all triangles that share this vertex

	if (do_smoothing) {
        int k, l, ci, nn;
        F_TYPE *cn=NULL, *on=NULL;
		triangle_list.start();
		for (i=0; i<triangle_list.n_items(); i++) {  // loop through triangles
			tri=triangle_list.current();

            // allocate new space for vertex normals
            tri->n1=new F_TYPE[3]; allocatedNormals.add(tri->n1);
            tri->n2=new F_TYPE[3]; allocatedNormals.add(tri->n2);
            tri->n3=new F_TYPE[3]; allocatedNormals.add(tri->n3);
            if (!tri->n1 || !tri->n2 || !tri->n3) MEMOUT;

            for (j=0; j<3; j++) {           // loop through vertexes
                if (j==0) { cn=tri->n1; ci=tri->i1; }
				else if (j==1) { cn=tri->n2; ci=tri->i2; }
				else { cn=tri->n3; ci=tri->i3; }

                for (k=0; k<3; k++) cn[k]=tri->fn[k];   // start ave. with current face normal
                nn=1;

				for (k=0; k<(shared_vertexes[ci]->n_items()); k++) {

					//
                    // and average values if angle between face normals < creaseAngle. Calculate
					// dot product of normals: if dot product (=cos(angle)) > cos(creaseAngle)
					// then angle < creaseAngle. (Saves having to do (slow) cos calculation
					// on every single pair of normals...)
					//

					if ((shared_vertexes[ci]->current())!=tri) {
                        on=shared_vertexes[ci]->current()->fn;
                        F_TYPE dotp=tri->fn[0]*on[0]+tri->fn[1]*on[1]+tri->fn[2]*on[2];
						if (dotp>cos_creaseAngle) {
							for (l=0; l<3; l++) cn[l]+=on[l];
							nn++;
						}
					}
					shared_vertexes[ci]->next();
				}
                if (nn>1) for (l=0; l<3; l++) cn[l]/=((F_TYPE)nn);
			}
			triangle_list.next();
		}
    }

	// print out the mesh

    // do colors and count unculled triangles
    int nUnculledTriangles=0;
    triangle_list.start();
    for (i=0; i<triangle_list.n_items(); i++) {
        tri=triangle_list.current();
        if (!tri->culled) {
            nUnculledTriangles++;
            if (colors) {
                out << "#declare t_" << i << " = texture { pigment { rgb";
                if (use_transparency) out << 't';
                out << " <" << tri->color[0] << ',' << tri->color[1] << ','
                    << tri->color[2];
                if (use_transparency) out << ',' << transparency;
                out << "> }";
                char* fin=strstr(finish,"finish");
                if (fin) out << ' ' << fin;
                out << " }\n";
            }
        }
        triangle_list.next();
    }
    if (nUnculledTriangles) {
        out << INDENT << "mesh {\n";
        indent+=ii;
        triangle_list.start();
        for (i=0; i<triangle_list.n_items(); i++) {
            tri=triangle_list.current();
            if (!tri->culled) {
                if (smooth_tri) {
                    out << INDENT << "smooth_triangle {\n";
                    indent+=ii;
                    out << INDENT << '<'
                        << coords[3*(tri->i1)] << ',' << coords[3*(tri->i1)+1] << ','
                        << coords[3*(tri->i1)+2] << ">, <"
                        << tri->n1[0] << ',' << tri->n1[1] << ',' << tri->n1[2] << ">,\n"
                        << INDENT << '<'
                        << coords[3*(tri->i2)] << ',' << coords[3*(tri->i2)+1] << ','
                        << coords[3*(tri->i2)+2] << ">, <"
                        << tri->n2[0] << ',' << tri->n2[1] << ',' << tri->n2[2] << ">,\n"
                        << INDENT << '<'
                        << coords[3*(tri->i3)] << ',' << coords[3*(tri->i3)+1] << ','
                        << coords[3*(tri->i3)+2] << ">, <"
                        << tri->n3[0] << ',' << tri->n3[1] << ',' << tri->n3[2] << ">\n";
                } else {
                    out << INDENT << "triangle {\n";
                    indent+=ii;
                    out << INDENT << '<'
                        << coords[3*(tri->i1)] << ',' << coords[3*(tri->i1)+1] << ','
                        << coords[3*(tri->i1)+2] << ">, <"
                        << coords[3*(tri->i2)] << ',' << coords[3*(tri->i2)+1] << ','
                        << coords[3*(tri->i2)+2] << ">, <"
                        << coords[3*(tri->i3)] << ',' << coords[3*(tri->i3)+1] << ','
                        << coords[3*(tri->i3)+2] << ">\n";
                }
                if (colors) {
                    out << INDENT << "texture { t_" << i << " }\n";
                }
                indent-=ii;
                out << INDENT << "}\n";
            }
            triangle_list.next();
        }
        out << INDENT << current_transformation << '\n';
        if (!colors) {
            if (finish[0]) out << INDENT << finish << '\n';
        }
        indent-=ii;
        out << INDENT << "}\n";
    }

	// destroy triangles and empty the list
    triangle_list.empty_and_delete_all();
    allocatedNormals.empty_and_delete_all();

	if (do_smoothing) {
		for (i=0; i<n_coords; i++) {
            delete shared_vertexes[i];
		}
		delete shared_vertexes;
	}

	return;
}
