/*
* line_pnt.cpp
*
* $Id: line_pnt.cpp,v 1.4 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* This module contains IndexedLineSet and PointSet translators.
*/

#include "pov.h"
#include "povout.h"

using namespace std;

#define PROBLEM(WHAT) do { \
	cerr << "Yikes! Problem with IndexedLineSet: " \
	<< WHAT << '\n'; \
	return; \
} while (0)

inline int is_length_zero(F_TYPE *v)   // allow for small digital error
{
	return ((v[0]*v[0]+v[1]*v[1]+v[2]*v[2]) < (1.0e-12));
}

inline int is_same_point(F_TYPE *v1, F_TYPE *v2)
{
	return (fabs((v1[0]-v2[0])) < (1.0e-3) &&
		fabs((v1[1]-v2[1])) < (1.0e-3) &&
		fabs((v1[2]-v2[2])) < (1.0e-3));
}

void do_IndexedLineSet(SFNode* node)
{
	F_TYPE* coords=NULL;
	int n_coords=0;
	SFNode* tmp=FIND(node,coord,SFNode);
	if (tmp) {
		coords=(FIND(tmp,point,MFVec3f))->value;
		n_coords=(FIND(tmp,point,MFVec3f))->n_items;
	}
	if (n_coords<2 || !coords) {
		cerr << "Error in IndexedLineSet - can't have < 2 coordinates!\n";
		return;
	}
	
	F_TYPE* colors=NULL;
	int n_colors=0;
	tmp=FIND(node,color,SFNode);
	if (tmp) {
		colors=(FIND(tmp,color,MFColor))->value;
		n_colors=(FIND(tmp,color,MFColor))->n_items;
	}
	
	I_TYPE* coordIndex=(FIND(node,coordIndex,MFInt32))->value;
	int n_coordIndexes=(FIND(node,coordIndex,MFInt32))->n_items;
	if (n_coordIndexes<2) {
		cerr << "Error in IndexedLineSet - can't have < 2 coordIndexes!\n";
		return;
	}
	
	I_TYPE* colorIndex=(FIND(node,colorIndex,MFInt32))->value;
	int n_colorIndexes=(FIND(node,colorIndex,MFInt32))->n_items;
	
	int colorPerVertex=FIND(node,colorPerVertex,SFBool)->value;
	
	// output a "line radius" declaration so line thickness can be easily changed
	
    out << "#declare LR=" << line_radius << ";\n"
		<< INDENT << "union {\n";
	indent+=ii;
	
	// loop through index, and output cylinder for each segment
	
	int i1, i2;
	int p1=0, p2;
	F_TYPE *c1, *c2;
	int n_lines=0;
	do {
		p2=p1+1;
		i1=coordIndex[p1];
		i2=coordIndex[p2];
		if (i1!=-1 && i2!=-1) {
			if (i1>=n_coords || i2>=n_coords) PROBLEM("not enough coords");
			c1=&(coords[3*i1]);     // get coordinates for segment ends
			c2=&(coords[3*i2]);
			
			// determine the color of each end
			
			F_TYPE *color1, *color2;
			if (colors) {
				if (colorPerVertex) {
					if (n_colorIndexes>0) {
						if (p1>=n_colorIndexes || p2>=n_colorIndexes)
							PROBLEM("not enough colorIndexes");
						if (colorIndex[p1]>=n_colors || colorIndex[p2]>=n_colors)
							PROBLEM("color index out of range");
						color1=&(colors[3*colorIndex[p1]]);
						color2=&(colors[3*colorIndex[p2]]);
					} else {
						if (coordIndex[p1]>=n_colors || coordIndex[p2]>=n_colors)
							PROBLEM("color index out of range");
						color1=&(colors[3*coordIndex[p1]]);
						color2=&(colors[3*coordIndex[p2]]);
					}
				} else {
					if (n_colorIndexes>0) {
						if (n_lines>=n_colorIndexes) PROBLEM("not enough colorIndexes");
						if (colorIndex[n_lines]>=n_colors) PROBLEM("colorIndex out of range");
						color1=color2=&(colors[3*colorIndex[n_lines]]);
					} else {
						if (n_lines>=n_colors) PROBLEM("not enough colors");
						color1=color2=&(colors[3*n_lines]);
					}
				}
			} else {
				if (!emissiveColor) PROBLEM("color=NULL and emissiveColor not set");
				color1=color2=emissiveColor->value;
			}
			
			// output the cylinder...
			
			if (is_same_point(color1,color2) && !is_same_point(c1,c2)) {
				
				// ... pretty easy if ends are the same color
				
				out << INDENT << "cylinder {\n";
				indent+=ii;
				out << INDENT << '<' << c1[0] << ',' << c1[1] << ',' << c1[2]
					<< ">, <" << c2[0] << ',' << c2[1] << ',' << c2[2] << ">, LR\n"
					<< INDENT << "pigment { rgb <"
					<< color1[0] << ',' << color1[1] << ',' << color1[2]
					<< "> } finish { ambient 0.6 }\n";
				indent-=ii;
				out << INDENT << "}\n";
				
			} else {
				
				//
				// ... but considerably more difficult if they're not. To shade smoothly
				// along a cylinder, POV requires it to be on a principal axis, so create
				// the cylinder on the +Y axis, then scale/rotate/translate it into place.
				//
				
				out << INDENT << "cylinder {\n";
				indent+=ii;
				out << INDENT << "<0,0,0>, <0,1,0>, LR\n"
					<< INDENT << "pigment { gradient y\n"
					<< INDENT << "  color_map { [0.0 color rgb <"
					<< color1[0] << ',' << color1[1] << ',' << color1[2]
					<< "> ] [ 1.0 color rgb <"
					<< color2[0] << ',' << color2[1] << ',' << color2[2]
					<< "> ] }\n" << INDENT << "} finish { ambient 0.6 }\n";
				
				static Matrix m1, m2, m3;
				static SFVec3f vec;
				vec.value[0]=vec.value[2]=1.0;
				vec.value[1]=sqrt((c2[0]-c1[0])*(c2[0]-c1[0])+
					(c2[1]-c1[1])*(c2[1]-c1[1])+
					(c2[2]-c1[2])*(c2[2]-c1[2]));
				set_scale_matrix(&m1,&vec); // scale cylinder to right length
				
				vec.value[0]=(c2[0]-c1[0])/vec.value[1];
				vec.value[2]=(c2[2]-c1[2])/vec.value[1]; // unit vector from c1 to c2
				vec.value[1]=(c2[1]-c1[1])/vec.value[1];
				
                SFRotation *rot = new SFRotation();
                rot->value[0]=vec.value[2];  // simplified rot=(Y axis) cross vec
                rot->value[1]=0.0;           // used to rotate cylinder into position
                rot->value[2]=-vec.value[0];
                if (is_length_zero(rot->value)) {  // vec along Y axis
                    rot->value[0]=rot->value[1]=0.0;
                    rot->value[2]=1.0;
					if (vec.value[1]<0.0)
                        rot->value[3]=3.14159265358979323846;  // -Y: rotate 180 deg.
					else
                        rot->value[3]=0.0;                     // +Y
				} else {
                    rot->value[3]=acos(vec.value[1]); // rotate cylinder onto vec
				}
				set_rotation_matrix(&m2,rot);
				compose_into(&m2,&m1,&m3);  // first scale, then rotate...
				
				vec.value[0]=c1[0];
				vec.value[1]=c1[1];
				vec.value[2]=c1[2];
				set_translation_matrix(&m1,&vec);
				compose_into(&m1,&m3,&m2);  // ... then translate ...
				
				out << INDENT << &m2 << '\n';  // ... the cylinder into place.
				indent-=ii;
				out << INDENT << "}\n";
			}
			
    } else if (i2==-1) n_lines++;
    p1++;
  } while (p1<(n_coordIndexes-1));
  
  out << INDENT << current_transformation << '\n';
  indent-=ii;
  out << INDENT << "}\n";
  return;
}

void do_PointSet(SFNode* node)
{
	F_TYPE* coords=NULL;
	int n_coords=0;
	SFNode* tmp=FIND(node,coord,SFNode);
	if (tmp) {
		coords=(FIND(tmp,point,MFVec3f))->value;
		n_coords=(FIND(tmp,point,MFVec3f))->n_items;
	}
	if (!coords || n_coords<1) return;
	
	F_TYPE* colors=NULL;
	int n_colors=0;
	tmp=FIND(node,color,SFNode);
	if (tmp) {
		colors=(FIND(tmp,color,MFColor))->value;
		n_colors=(FIND(tmp,color,MFColor))->n_items;
	}
	
	if (colors && n_colors<n_coords) {
		cerr << "Error in PointSet: too few colors.\n";
		return;
	}
	if (!colors && !emissiveColor) {
		cerr << "Error in PointSet: no colors, and no emissiveColor set\n";
		return;
	}
	
	// output a "point radius" declaration so radius can be easily changed
	
    out << "#declare PR=" << point_radius << ";\n"
		<< INDENT << "union {\n";
	indent+=ii;
	
	for (int i=0; i<n_coords; i++) {
		out << INDENT << "sphere { <"
			<< coords[3*i] << ',' << coords[3*i+1] << ',' << coords[3*i+2]
			<< ">, PR pigment { rgb <";
		if (colors)
			out << colors[3*i] << ',' << colors[3*i+1] << ',' << colors[3*i+2];
		else
			out << emissiveColor->value[0] << ',' << emissiveColor->value[1]
			<< ',' << emissiveColor->value[2];
		out << "> } finish { ambient 0.6 } }\n";
	}
	
	out << INDENT << current_transformation << '\n';
	indent-=ii;
	out << INDENT << "}\n";
	return;
}
