/*
* list.h
*
* $Id: list.h,v 1.5 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*            http://www.chemicalgraphics.com/paul
*
* Defines the DLList class, a general template-based class that allows
* creation and manipulation of (doubly linked) lists of any type object.
*/

#ifndef __vrml2pov_list_h_
#define __vrml2pov_list_h_

#include <stdlib.h>

//
// A general list template to create doubly linked lists of any type object.
// The list is, for simplicity of implementation, actually circular, but
// there is always a pointer to the "first" element of the list, a count
// of how many items the list contains, and a "current" item (mostly to
// be used to move through the list). When a new item is are put into the list,
// that new item always becomes the current item.
//
// List functions available to the user:
//   - first(), last(), current() retrieve the appropriate item
//   - n_items() returns the count
//   - insert(item) puts the new item *right after* the "current" item
//     and then this new items becomes the current item
//   - add(item) adds the new item to the end of the list and makes
//     it current
//   - prepend(item) puts the new item at the front of the list, making
//     it both "first" and current
//   - next() and prev() move the current pointer forward or back
//   - remove() removes the current item from the list, sets "current" item
//     to previous item
//   - start() make first() the current()
//   - empty() empties the list (but does not destroy contained objects!)
//
// Note that items are held by pointers, and only the pointer is actually
// stored. So, items are not copied onto the list, nor are they ever changed
// by list functions or destroyed when remove()'ed.
//

template<class T>
class DLList;

template<class T>
class TLink {
private:
	friend class DLList<T>;
	T* pntr;
	TLink<T> *pr, *nx;
	TLink(T* p, TLink<T> *prv, TLink<T> *nxt)
	{
		pntr=p; pr=prv; nx=nxt;
	}
	TLink(T* p)
	{
		pntr=p; pr=nx=this;
	}
};

template<class T>
class DLList {
private:
	TLink<T> *f, *cur;
	int count;
public:
	DLList() { count=0; f=cur=NULL; }
	~DLList() { empty(); }
	T* first() { if (count) return f->pntr; else return NULL; }
	T* last() { if (count) return f->pr->pntr; else return NULL; }
	T* current() { if (count) return cur->pntr; else return NULL; }
	int n_items() { return count; }
	void insert(T*);
	void prepend(T*);
	void add(T*);
	void remove();
	void start() { cur=f; }
	void next();
	void prev();
	void empty();
    void empty_and_delete_all();
};

template<class T>
inline void DLList<T>::insert(T* n)
{
	if (count)
		cur=cur->nx=cur->nx->pr=new TLink<T>(n,cur,cur->nx);
	else
		f=cur=new TLink<T>(n);
	if (cur==NULL) {
		std::cerr << "Out of memory during TLink allocation\n";
        abort();
	}
	count++;
}

template<class T>
inline void DLList<T>::prepend(T* n)
{
	if (count)
		cur=f=f->pr=f->pr->nx=new TLink<T>(n,f->pr,f);
	else
		f=cur=new TLink<T>(n);
	if (cur==NULL) {
		std::cerr << "Out of memory during TLink allocation\n";
        abort();
	}
	count++;
}

template<class T>
inline void DLList<T>::add(T* n)
{
	if (count)
		cur=f->pr=f->pr->nx=new TLink<T>(n,f->pr,f);
	else
		f=cur=new TLink<T>(n);
	if (cur==NULL) {
		std::cerr << "Out of memory during TLink allocation\n";
        abort();
	}
	count++;
}

template<class T>
inline void DLList<T>::remove()
{
	TLink<T> *tmp;
	if (count) {
		cur->nx->pr=cur->pr;
		cur->pr->nx=cur->nx;
		if (f==cur) f=cur->nx;
        tmp=cur->pr;    // set pointer to *previous* item (for typical use as stack)
		delete cur;
		count--;
		if (count)
			cur=tmp;
		else
			cur=f=NULL;
	} else
		std::cerr << "Warning! Tried to remove() from an empty list!\n";
}

template<class T>
inline void DLList<T>::next()
{
	if (count) cur=cur->nx;
}

template<class T>
inline void DLList<T>::prev()
{
	if (count) cur=cur->pr;
}

template<class T>
inline void DLList<T>::empty()
{
	int i;
	TLink<T> *tmp;
	for (i=0;i<count;i++) {
		if (i<count-1) {
			tmp=cur->nx; delete cur; cur=tmp;
		} else
			delete cur;
	}
	count=0;
	cur=f=NULL;
}

template<class T>
inline void DLList<T>::empty_and_delete_all()
{
	int i;
	TLink<T> *tmp;
	for (i=0;i<count;i++) {
        delete cur->pntr;
		if (i<count-1) {
			tmp=cur->nx; delete cur; cur=tmp;
		} else
			delete cur;
	}
	count=0;
	cur=f=NULL;
}

#endif // __vrml2pov_list_h_
