/*
* node.cpp
*
* $Id: node.cpp,v 1.6 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*            http://www.chemicalgraphics.com/paul
*
* Implements the parsing of a node. Returns a pointer to the new node(s)
* if successful, NULL if not. These functions are the same as the
* rest of the parse_[SM]F* functions in that they expect current_word
* to be set to the first word of the node statement.
*/

#include <ctype.h>

#include "parser.h"
#include "node.h"
#include "stdnodes.h"

using namespace std;

DLList<fieldType> fieldType::allocatedNodes;

SFNode* create_node_from_PROTO(ParserObject*, SFNode*);
SFNode* create_node_from_internal_PROTO_def(ParserObject*,
											SFNode*,
											protoStatement*,
											DLList<nodeVariable> *);
SFNode* traverse_node(SFNode*,
					  SFNode*, DLList<nodeVariable> *);

fieldType* parse_SFNode(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience

	SFNode* node=new SFNode(NULL);
	if (!node) OUTMEM;

	// if USE statement, look for this node in the current context.

	int i;
	if (strcmp(r.current_word,"USE")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("nodeTypeId",NULL);
		for (i=0; i<(p->current_DEFs->n_items()); i++) {
			if (strcmp(r.current_word,p->current_DEFs->current()->nodeNameId)==0) {
				if (p->isVerbose) cout << "USEd " << p->current_DEFs->current()->nodeNameId << '\n';
				return p->current_DEFs->current()->node;
			}
			p->current_DEFs->next();
		}
		r.parse_error("nodeNameId",r.current_word);
		cerr << "Node " << r.current_word << " not defined in this context.\n";
		return NULL;
	}

	// Store DEF name in current_DEFs if found

	if (strcmp(r.current_word,"DEF")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("nodeNameId",NULL);
        if (!is_valid_id(r.current_word))
			return (fieldType*)r.parse_error("valid nodeNameId",r.current_word);
		nodeStatement* ns=new nodeStatement(r.current_word,node);
		if (!ns || !ns->nodeNameId) OUTMEM;
		p->current_DEFs->add(ns);
		if (p->isVerbose) cout << "DEF'ed " << r.current_word << '\n';
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("nodeTypeId",NULL);
	}

	//
	// Decide whether this is a standard node or a PROTO-based node. If the
	// latter, then create this node according to the PROTO. Otherwise, read
	// in the node's fields that are specified.
	//

	i=find_standard_node(r.current_word);
    if (i==0) {
        if (p->isVerbose) cout << r.current_word << " (PROTO-based) node found\n";
        return create_node_from_PROTO(p,node);
    }
	node->nodeTypeId=new char[strlen(r.current_word)+1];
	if (!node->nodeTypeId) OUTMEM;
	strcpy(node->nodeTypeId,r.current_word);
	if (i==UNIMPLEMENTED) {
		if (strcmp(r.current_word,"Script")==0) {
			cout << "Warning: file " << r.inputName << " line " << r.current_line
				<< " column " << r.current_pos-5
				<< ":\n  No scripting languages implemented; skipping Script { }\n";
		} else
			if (p->isVerbose) cout << r.current_word <<
				" (standard) node found, but is not deemed useful so will left empty\n";
			if (!skip_node(p))
				return NULL;
			else
				return node;
	} else
		if (p->isVerbose) cout << r.current_word << " (standard) node found\n";

		//
		// Here we branch parsing paradigms. If we're currently inside a PROTO
		// definition, the definition will be read in literally, including IS
		// statements, and standard nodes will not be filled out with all their
		// default values. If we're not in a PROTO, then this is an actual
		// instantiation of this node, and it will be filled in afterwards with default
		// values for all its fields if they are not specified in the node statement.
		//

		// after typeId, expect '{' then first word of nodeBodyElement

		if (!r.get_next_word())
			return (SFNode*)r.parse_error("{",NULL);
		if (strcmp(r.current_word,"{")!=0)
			return (SFNode*)r.parse_error("{",r.current_word);
		if (!r.get_next_word())
			return (SFNode*)r.parse_error("nodeBodyElement or }",NULL);

		// read the nodeBodyElements and deal accordingly...

		while (strcmp(r.current_word,"}")!=0) {

			if (strcmp(r.current_word,"PROTO")==0) {
				if (p->isVerbose) cout << "Parsing PROTO ";
				protoStatement* new_PROTO=parse_PROTO(p);
				if (!new_PROTO) {
					cerr << "Error in PROTO statement\n";
					return NULL;
				} else
                    p->local_PROTOs->add(new_PROTO);
			}

			// skip EXTERNPROTO with warning (eventually?)

			else if (strcmp(r.current_word,"EXTERNPROTO")==0) {
				cerr << r.current_word << "currently unimplemented... sorry.\n";
				return NULL;
			}

			// ignore ROUTEs

			else if (strcmp(r.current_word,"ROUTE")==0) {
				if (!skip_route(p)) {
					cerr << "Error in ROUTE statement\n";
					return NULL;
				}
			}

			// All else is assumed to be a nodeBodyElement; first get id and type info

			static char id[BUFFERSIZE];
			strcpy(id,r.current_word);
			int type=node_field_type(node->nodeTypeId,id);
			if (!type) {
				r.parse_error("fieldId",r.current_word);
				cerr << "Field \"" << id << "\" is not a known field of "
					<< node->nodeTypeId << '\n';
				return NULL;
			}
			if (!r.get_next_word())
				return (SFNode*)r.parse_error("fieldValue or IS",NULL);

			// If we're in a PROTO, add IS declarations to nodeBodyArgs

			if (strcmp(r.current_word,"IS")==0) {
				if (!p->in_PROTO)
					return (SFNode*)r.parse_error("fieldValue","IS outside PROTO");
				if (!r.get_next_word())
					return (SFNode*)r.parse_error("fieldValue",NULL);

				// make sure this is a valid argument, and of correct type

				for (i=0; i<(p->arg_Stack.current()->n_items()); i++) {
					if (strcmp(p->arg_Stack.current()->current()->fieldId,
						r.current_word)==0) {
						p->arg_Stack.current()->current()->used=TRUE;
						break;
					}
					p->arg_Stack.current()->next();
				}
				if (i==p->arg_Stack.current()->n_items()) {
					r.parse_error("PROTO argument name",r.current_word);
					cerr << "argument " << r.current_word <<
						" is not one of this PROTO's arguments!\n";
					return NULL;
				}
				if (p->arg_Stack.current()->current()->defaultValue &&
					type!=(p->arg_Stack.current()->current()
					->defaultValue->fieldTypeName)) {
					r.parse_error(p->type_name[type],r.current_word);
					cerr << "type mismatch: cannot assign " << r.current_word <<
						" to " << id << '\n';
					return NULL;
				}

				// add to arguments list

				if (type!=EVENT) { // ignore events
					nodeISarg* arg=new nodeISarg(id,r.current_word);
					if (!arg) OUTMEM;
					if (!node->nodeBodyArgs) {
						node->nodeBodyArgs=new DLList<nodeISarg>;
						if (!node->nodeBodyArgs) OUTMEM;
					}
					node->nodeBodyArgs->add(arg);
				}
			} // if IS statement

			// add variable to nodeBody

			else {
				if (type!=EVENT) {
					nodeVariable* var=new nodeVariable(id,NULL);
					if (!var || !var->fieldId) OUTMEM;
					if (type==SFNODE && strcmp(r.current_word,"NULL")==0)
						var->fieldValue=NULL;
					else {
						var->fieldValue=(*p->parse_type[type])(p);
						if (!var->fieldValue) {
							cerr << "Error parsing type " << p->type_name[type] << '\n';
							delete var;
							return NULL;
						}
					}
					node->nodeBody.add(var);
				}
			}

			if (!r.get_next_word())
				return (SFNode*)r.parse_error("nodeBodyStatement or }",NULL);
  } // while reading nodeBodyElements

  //
  // For the sake of efficiency, only fill in the node with defaults if this
  // is an actual instantiation of the standard node - i.e., if we're not
  // inside a PROTO statement. However, do fill it out if the node is being
  // used as a default value for a PROTO argument.
  //

  if (!p->in_PROTO || p->in_PROTO_declarations) {
	  if (!finish_node(node)) {
		  cerr << "Internal error finishing node " << node->nodeTypeId << '\n';
		  return NULL;
	  }

	  //
	  // When an Inline node is instantiated, add nodes read from the URL
	  // to a new MFNode called "children".

	  if (strcmp(node->nodeTypeId,"Inline")==0) {

		  // find url

		  MFString* url=NULL;
		  for (i=0; i<(node->nodeBody.n_items()); i++) {
			  if (strcmp(node->nodeBody.current()->fieldId,"url")==0) {
				  url=(MFString*)(node->nodeBody.current()->fieldValue);
				  break;
			  }
			  node->nodeBody.next();
		  }
		  if (i==node->nodeBody.n_items() ||
			  !url || url->fieldTypeName!=MFSTRING) {
			  cerr << "Internal error: url field of Inline not found!\n";
			  return NULL;
		  }

		  // read first valid file found; add nodes to inline_nodes

		  static DLList<SFNode> inline_nodes;
		  i=0;
		  while (i<url->n_items && url->value[i]) {
			  if (p->isVerbose) { cout << "Reading inline file " <<
				  url->value[i] << ":\n"; }
			  p->push_buffer(url->value[i],&inline_nodes);
			  if (p->current_file->inputName[0] != '\0') {
				  if (parse_file(p)) {
					  cerr << "Error in inline file " << url->value[i] << '\n';
					  return NULL;
				  } else {
					  p->pop_buffer();
					  break;
				  }
			  } else
				  i++;	// couldn't open file; try next one.
			  p->pop_buffer();
		  } // while reading MFString

		  // create new nodeVariable - an MFNode into which is copied the inline_nodes

		  nodeVariable* var=new nodeVariable("children",NULL);
		  if (!var || !var->fieldId) OUTMEM;
		  var->fieldValue=new MFNode(inline_nodes.n_items());
		  if (!var->fieldValue) OUTMEM;
		  inline_nodes.start();
		  for (i=0; i<inline_nodes.n_items(); i++) {
			  ((MFNode*)var->fieldValue)->value[i]=inline_nodes.current();
			  inline_nodes.next();
		  }
		  node->nodeBody.add(var);
		  inline_nodes.empty();

	  }
  }
  return node;
}

//
// Read in an MFNode. In this case we have to be careful, because this
// routine could also be called recursively (if a node it parses contains
// an MFNode field), so we can't use a static buffer list. Create a new
// one each time.
//

fieldType* parse_MFNode(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience

	MFNode* mfn=NULL;
	SFNode* sfn=NULL;

	// single node

	if (strcmp(r.current_word,"[")!=0) {
		sfn=(SFNode*)parse_SFNode(p);
		if (!sfn) {
			cerr << "Error in node statement\n";
			return NULL;
		}
		mfn=new MFNode(1);
		if (!mfn) OUTMEM;
		mfn->value[0]=sfn;
	}

	// multiple nodes

	else {
		DLList<SFNode> buffer;
		if (!r.get_next_word())
			return (SFNode*)r.parse_error("nodeTypeId",NULL);
		while (strcmp(r.current_word,"]")!=0) {
			sfn=(SFNode*)parse_SFNode(p);
			if (!sfn) {
				cerr << "Error in node statement\n";
				return NULL;
			}
			buffer.add(sfn);
			if (!r.get_next_word())
				return (SFNode*)r.parse_error("nodeTypeId or ]",NULL);
		}
		mfn=new MFNode(buffer.n_items());
		if (!mfn) OUTMEM;
		buffer.start();
		for (int i=0; i<buffer.n_items(); i++) {
			mfn->value[i]=buffer.current();
			buffer.next();
		}
	}

	return mfn;
}

//
// Go past node up to closing }, taking care not to count { or } inside
// strings... (what a pain.) Return 0 on failure.
//

int skip_node(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience

	if (!r.get_next_word()) {
		cerr << "Did not find { for node\n";
		return 0;
	}
	if (strcmp(r.current_word,"{")!=0) {
		cerr << "Internal error - skip_node expected '{'\n";
        abort();
	}

	char ch;
	int nest_level=1;
	int in_string=FALSE;
	while (nest_level!=0) {
		ch=r.get_ch();
		if (!ch) {
			cerr << "Did not reach end of node\n";
			return 0;
		}
		switch (ch) {
		case '{':
			if (!in_string) nest_level++;
			break;
		case '}':
			if (!in_string) nest_level--;
			break;
		case '"':
			in_string=!in_string;
			break;
		case '\\':
			r.get_ch(); // skip next character as literal
			break;
		default:
			break;
		}
	}
	return 1;
}

//
// Create a node out of a PROTO definition.
//

SFNode* create_node_from_PROTO(ParserObject* p, SFNode* node)
{
	ReadFileObject& r=*(p->current_file); // for convenience

	// figure out what PROTO we're using.
    protoStatement* proto=p->FindPROTO(r.current_word, &r);
    if (!proto) return NULL;

	//
	// The first step is to create a list of all arguments. First check to see
	// that the argument passed is actually part of the PROTO's argument list.
	// If so, create a new field of that type to use in the expansion of the
	// PROTO into a node. If an argument is in the PROTO's argument list, but
	// not present in the call to this PROTO, then simply link in a pointer to
	// the defaultValue.
	//

	if (!r.get_next_word())
		return (SFNode*)r.parse_error("{",NULL);
	if (strcmp(r.current_word,"{")!=0)
		return (SFNode*)r.parse_error("{",r.current_word);
	if (!r.get_next_word())
		return (SFNode*)r.parse_error("fieldId or }",NULL);

	DLList<nodeVariable> arguments;
    int type=ERROR, i;
	for (i=0; i<(proto->interfaceDeclarations.n_items()); i++) {
		proto->interfaceDeclarations.current()->used=FALSE;
		proto->interfaceDeclarations.next();
	}

	while (strcmp(r.current_word,"}")!=0) {

		// Look for arg in arg list, both to check if present and to get type

		for (i=0; i<(proto->interfaceDeclarations.n_items()); i++) {
			if (strcmp(proto->interfaceDeclarations.current()->fieldId,
				r.current_word)==0) {
				if (!(proto->interfaceDeclarations.current()->defaultValue))
					type=SFNODE;	// NULL SFNode argument - can't get type directly
				else {
					type=proto->interfaceDeclarations.current()
						->defaultValue->fieldTypeName;
					if (!type) {
						cerr << "Ack! Internal error! Somehow picked up an invalid type.\n";
                        abort();
					}
				}
				break;
			}
			proto->interfaceDeclarations.next();
		}
		if (i==(proto->interfaceDeclarations.n_items())) {
			r.parse_error("argument",r.current_word);
			cerr << "field " << r.current_word <<
				" is not an argument for " << proto->nodeTypeId << "!\n";
			return NULL;
		}
		proto->interfaceDeclarations.current()->used=TRUE;

		// get and store argument value (except events - skip those)

		if (!r.get_next_word())
			return (SFNode*)r.parse_error((type>0)?p->type_name[type]:"eventId",NULL);
		if (type!=EVENT) {
			nodeVariable* var=new
				nodeVariable(proto->interfaceDeclarations.current()->fieldId,NULL);
			if (!var || !(var->fieldId)) OUTMEM;
			if (type==SFNODE && strcmp(r.current_word,"NULL")==0)
				var->fieldValue=NULL;
			else {
				var->fieldValue=((*p->parse_type[type])(p));
				if (!var->fieldValue) {
					cerr << "Error parsing type " << p->type_name[type] << '\n';
					return NULL;
				}
			}
			arguments.add(var);
		}

		if (!r.get_next_word())
			return (SFNode*)r.parse_error("fieldId or }",NULL);
	} // while reading arguments

	// End of PROTO call; add pointer to default values if arguments not present

	for (i=0; i<(proto->interfaceDeclarations.n_items()); i++) {
		if (!(proto->interfaceDeclarations.current()->used)) {
			nodeVariable* var=new
				nodeVariable(proto->interfaceDeclarations.current()->fieldId,NULL);
			if (!var || !(var->fieldId)) OUTMEM;
			var->fieldValue=proto->interfaceDeclarations.current()->defaultValue;
			arguments.add(var);
		}
		proto->interfaceDeclarations.next();
	}

	// actually create the node using these arguments.

	node=create_node_from_internal_PROTO_def(p,node,proto,&arguments);
	if (!node) {
		cerr << "Internal error creating node!\n";
		return NULL;
	}

	arguments.empty_and_delete_all();
	return node;
}

//
// By definition, the PROTO's first node is the *only* node created by a call
// to a PROTO! Descend into this node (recursively, if it contains children)
// reading the definition and instantiating the new node.
//

SFNode* create_node_from_internal_PROTO_def(ParserObject* p,
											SFNode* node,
											protoStatement* proto,
											DLList<nodeVariable> *arguments)
{
	p->push_context(proto);
	SFNode* protonode=proto->local_Nodes.first();

	// is this standard or user node?

	int node_status=find_standard_node(protonode->nodeTypeId);
	if (!node_status) {

		// user (nested PROTO) node

        protoStatement* nestedproto=p->FindPROTO(protonode->nodeTypeId, NULL);
        if (nestedproto) {
            if (p->isVerbose) cout << "creating nested PROTO-base node"
                << nestedproto->nodeTypeId << '\n';
            return create_node_from_internal_PROTO_def(p,node,
                nestedproto,
                arguments);
        }
        else {
			return NULL;
		}
	}

	// descend into standard node and create it and recursively create its children

	node->nodeTypeId=new char[strlen(protonode->nodeTypeId)+1];
	if (!node->nodeTypeId) OUTMEM;
	strcpy(node->nodeTypeId,protonode->nodeTypeId);
	if (p->isVerbose) cout << "Instantiating " << protonode->nodeTypeId
		<< " node based on PROTO "
		<< proto->nodeTypeId << '\n';

	if (node_status==UNIMPLEMENTED) {
		if (p->isVerbose)
			cout << "...but is not deemed useful so will be left empty\n";
	} else {
		node=traverse_node(node,protonode,arguments);
		if (!node) {
			cerr << "Internal error in node instantiation\n";
			return NULL;
		}
	}

	p->pop_context();
	return node;
}

//
// This routine descends into the node's structure (including recursion into
// children), filling it out by (essentially) copying the "protonode"
// structure, filling in "IS" assigned variables.
//

SFNode* traverse_node(SFNode* node,
					  SFNode* protonode, DLList<nodeVariable> *arguments)
{
	int i, j;
	nodeVariable* var;

	// assign values based on IS assignments

	if (protonode->nodeBodyArgs) {

		// for each IS assignment...

		for (i=0; i<(protonode->nodeBodyArgs->n_items()); i++) {
			var=new nodeVariable(protonode->nodeBodyArgs->current()->fieldId,NULL);
			if (!var || !var->fieldId) OUTMEM;

			// find passed argument to match...

			for (j=0; j<(arguments->n_items()); j++) {
				if (strcmp(arguments->current()->fieldId,
					protonode->nodeBodyArgs->current()->argId)==0) {
					var->fieldValue=arguments->current()->fieldValue;
					node->nodeBody.add(var);
					//	    if (var->fieldValue->fieldTypeName==SFNODE)
					//	      if (!finish_node((SFNode*)(var->fieldValue))) {
					//		cerr << "Internal error finishing node "
					//		     << ((SFNode*)(var->fieldValue))->nodeTypeId << '\n';
					//		return NULL;
					//	      }
					break;
				}
				arguments->next();
			}
			if (j==arguments->n_items()) {
				cerr << "Internal error: argument " << var->fieldId <<
					" not found in passed argument list!\n";
				return NULL;
			}

			protonode->nodeBodyArgs->next();
		}
	}

	// add regular fields

	protonode->nodeBody.start();
	for (i=0; i<(protonode->nodeBody.n_items()); i++) {
		var=new nodeVariable(protonode->nodeBody.current()->fieldId,NULL);
		if (!var || !var->fieldId) OUTMEM;

		// (recursively) create and finish SFNode field

		SFNode* subnode;
		nodeVariable* cpnode=protonode->nodeBody.current();

		if (cpnode->fieldValue==NULL)  // hope this is a NULL SFNode, or else...
			var->fieldValue=NULL;
		else if (cpnode->fieldValue->fieldTypeName==SFNODE) {
			subnode=new SFNode(((SFNode*)(cpnode->fieldValue))->nodeTypeId);
			if (!subnode || !(subnode->nodeTypeId)) OUTMEM;
			subnode=traverse_node(subnode,(SFNode*)(cpnode->fieldValue),arguments);
			if (!subnode) {
				cerr << "Internal error in SFNode instantiation\n";
				return NULL;
			}
			var->fieldValue=(fieldType*)subnode;
		}

		// (recursively) create and finish MFNode field

		else if (cpnode->fieldValue->fieldTypeName==MFNODE) {
			MFNode* mfnode=new MFNode(((MFNode*)(cpnode->fieldValue))->n_items);
			if (!mfnode) OUTMEM;
			for (j=0; j<(((MFNode*)(cpnode->fieldValue))->n_items); j++) {
				subnode=new
					SFNode(((MFNode*)(cpnode->fieldValue))->value[j]->nodeTypeId);
				if (!subnode || !(subnode->nodeTypeId)) OUTMEM;
				subnode=traverse_node(subnode,
					((MFNode*)(cpnode->fieldValue))->value[j],
					arguments);
				if (!subnode) {
					cerr << "Internal error in MFNode subnode instantiation\n";
					return NULL;
				}
				mfnode->value[j]=subnode;
			}
			var->fieldValue=(fieldType*)mfnode;
		}

		// regular field

		else {
			var->fieldValue=field_copy(cpnode->fieldValue);
		}
		node->nodeBody.add(var);
		protonode->nodeBody.next();
	}

	if (!finish_node(node)) {
		cerr << "Internal error finishing node " << node->nodeTypeId << '\n';
		return NULL;
	}
	return node;
}
