/*
* parser.cpp
*
* $Id: parser.cpp,v 1.6 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* The core of the parser.
*/

#include <stdio.h>
#include <iostream>
#include <fstream>

#include "parser.h"

using namespace std;

//
// This function should only be called once, to initiate parsing. It is
// the "front end" for the parser that gets called by the program that
// wants to deal with the VRML scene (in this case, vrml2pov).
//

int ParserObject::parse(const char* name)
{
	// first open the user's file

	current_file=new ReadFileObject(name);
	if (!current_file) {
		cerr << "No memory!\n";
		return 1;
	}

	// then read the user VRML file

	if (isVerbose) cout << "Reading file " << name << ":\n";
	if (parse_file(this)) {
		cerr << "Error in file " << name << '\n';
		return 1;
	}

	delete current_file;
	return 0; // Hooray! done without errors.
}

protoStatement* ParserObject::FindPROTO(const char *name, ReadFileObject *r)
{
    int i, j;
 
    // check local PROTO list first
    for (i=0; i<(local_PROTOs->n_items()); i++) {
        if (strcmp(name,local_PROTOs->current()->nodeTypeId)==0)
            return local_PROTOs->current();
        local_PROTOs->next();
	}

    // if not found, search outer PROTO contexts; this assumes that
    // the in PROTO_Stack, the current item is the last (innermost) context
    for (i=0; i<(PROTO_Stack.n_items()); i++) {
        for (j=0; j<(PROTO_Stack.current()->n_items()); j++) {
            if (strcmp(name,PROTO_Stack.current()->current()->nodeTypeId)==0) {
                protoStatement *proto=PROTO_Stack.current()->current();
                // IMPORTANT: must return PROTO_Stack current pointer to last item
                PROTO_Stack.start();
                PROTO_Stack.prev();
                return proto;
            }
            PROTO_Stack.current()->next();
        }
        PROTO_Stack.prev(); // search contexts in reverse order (innermost first)
    }

    // not found - error
    if (r) r->parse_error("PROTO name",name);
    cerr << "PROTO " << name << " undefined in current context!\n";
    return NULL;
}

//
// This routine parses a VRML file. It may also be used to parse inline
// (i.e. nested) input files, with calls to push_buffer and pop_buffer
// before and after, respectively - although this is not really necessary
// if Inline's are to be read after user files... Oh well, I built the
// ability to nest input files in here, so I may as well leave it in.
//

int parse_file(ParserObject *p)
{
	ReadFileObject& r=*(p->current_file);  // just for convenience

	// check to make sure this is a utf8-encoded VRML file

	const char* header="#VRML V2.0 utf8";  // strlen=15
	int i;

	for (i=0; i<15; i++) {
		if (r.get_ch() != header[i]) {
			cerr << r.inputName << ": Unknown filetype\n";
			return 1;
		}
	}
	if (p->isVerbose) cout << r.inputName << " is of correct type.\n";

	// read the rest of the first line, which is a comment by definition

	do {} while (!is_line_terminator(r.get_ch()));

	//
	// Now simply loop through the statements in the file, reading each one
	// with the appropriate type-specific parsing routine.
	//

	while (r.get_next_word()) {

		// read PROTO

		if (strcmp(r.current_word,"PROTO")==0) {
			if (p->isVerbose) cout << "Parsing PROTO ";
			protoStatement* new_PROTO=parse_PROTO(p);
			if (!new_PROTO) {
				cerr << "Error in PROTO statement\n";
				goto parse_error;
			} else
				p->local_PROTOs->add(new_PROTO);
		}

		// skip EXTERNPROTO with warning

		else if (strcmp(r.current_word,"EXTERNPROTO")==0) {
			cerr << r.current_word << "currently unimplemented... sorry.\n";
			goto parse_error;
		}

		// skip ROUTEs

		else if (strcmp(r.current_word,"ROUTE")==0) {
			if (!skip_route(p)) {
				cerr << "Error in ROUTE statement\n";
				goto parse_error;
			}
		}

		// everything else is assumed to be a node.

		else {
			SFNode* new_SFNode=(SFNode*)parse_SFNode(p);
			if (!new_SFNode) {
				cerr << "Error in node statement\n";
				goto parse_error;
			}
			p->current_Nodes->add(new_SFNode);
		}
	}

	r.in.close();
	if (p->isVerbose) cout << r.inputName << " parsed successfully.\n";
	return 0;

parse_error:
	r.in.close();
	return 1;
}

//
// Gets the next "word", defined here as something that is neither
// whitespace nor an "enclosure" character, i.e. { } [ ] "
// Also skips over comments.
//

int ReadFileObject::get_next_word(void)
{
	char ch;

	current_word[0]='\0';
	for (;;) {
		ch=peek_ch();
		if (!ch) return 0;
		if (is_ws(ch))
			get_ch();
		else if (ch=='#')
			do {} while (!is_line_terminator(get_ch()));
			else
				break;
	}
	for (int i=0; i<BUFFERSIZE; i++) {
		if (!(ch=peek_ch())) return 1;
		if (is_ws(ch) || is_enclosure(ch) || ch=='#') {
			if (i==0 && is_enclosure(ch)) {
				current_word[i++]=get_ch();
			}
			current_word[i]='\0';
			return 1;
		} else
			current_word[i]=get_ch();
	}
	cerr << "Word too long - increase BUFFERSIZE.\n";
	return 0;
}

void* ReadFileObject::parse_error(const char* e, const char* g)
{
	cerr << '\n' << inputName << ": Error at line " << current_line
		<< ", at or before column " << current_pos
		<< ": \nExpected " << e;
	if (g)
		cerr << ", but got " << g << "\n";
	else
		cerr << '\n';
	return NULL;
}

int skip_route(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file);  // just for convenience

	char ch;
	for (int i=0; i<2; ) {
		if ((ch=r.get_ch())=='.')
			i++;
		else if (ch=='\0') {
			cerr << "incomplete ROUTE\n";
			return 0;
		}
	}
	if (!r.get_next_word()) {
		cerr << "incomplete ROUTE\n";
		return 0;
	}
	return 1;
}

int is_valid_id(char *w)
{
    if ((*w>=0x30 && *w<=0x39) || (*w>=0x0 && *w<=0x20) || *w==0x22 ||
        *w==0x23 || *w==0x27 || *w==0x2b || *w==0x2c || *w==0x2d || *w==0x2e ||
        *w==0x5b || *w==0x5c || *w==0x5d || *w==0x7b || *w==0x7d || *w==0x7f)
        return FALSE;

    while (*(++w) != '\0') {
        if ((*w>=0x0 && *w<=0x20) || *w==0x22 || *w==0x23 || *w==0x27 ||
            *w==0x2c || *w==0x2e || *w==0x5b || *w==0x5c || *w==0x5d ||
            *w==0x7b || *w==0x7d || *w==0x7f)
            return FALSE;
    }

    return TRUE;
}

