/*
* parser.h
*
* $Id: parser.h,v 1.6 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*           http://www.chemicalgraphics.com/paul
*
* Header file for ParserObject.
*/

#ifndef __vrml2pov_parser_h_
#define __vrml2pov_parser_h_

#include <stdlib.h>
#include <iostream>
#include <fstream>

#include "global.h"
#include "fieldtyp.h"
#include "list.h"
#include "proto.h"
#include "node.h"

#define BUFFERSIZE 8192

//
// functions to parse and store particular fieldTypes - implemented in
// readtype.cpp, except SFNode and MFNode which are read by node.cpp.
//

class ParserObject;

fieldType* parse_SFBool(ParserObject*);
fieldType* parse_SFColor(ParserObject*);
fieldType* parse_MFColor(ParserObject*);
fieldType* parse_SFFloat(ParserObject*);
fieldType* parse_MFFloat(ParserObject*);
fieldType* parse_SFImage(ParserObject*);
fieldType* parse_SFInt32(ParserObject*);
fieldType* parse_MFInt32(ParserObject*);
fieldType* parse_SFRotation(ParserObject*);
fieldType* parse_MFRotation(ParserObject*);
fieldType* parse_SFString(ParserObject*);
fieldType* parse_MFString(ParserObject*);
fieldType* parse_SFTime(ParserObject*);
fieldType* parse_MFTime(ParserObject*);
fieldType* parse_SFVec2f(ParserObject*);
fieldType* parse_MFVec2f(ParserObject*);
fieldType* parse_SFVec3f(ParserObject*);
fieldType* parse_MFVec3f(ParserObject*);

class ReadFileObject {
public:
	char inputName[256];
	std::ifstream in;
	char buffer[BUFFERSIZE];
	int pos, n_chars;
	char current_word[BUFFERSIZE];
	int current_line, current_pos;
	char get_ch(void);
	char peek_ch(void);
	int get_next_word(void);
	void* parse_error(const char*, const char*);

	ReadFileObject(const char* file) {
		strcpy(inputName,file);
		in.open(file,std::ios::in);
		if (in.fail()) {
			std::cerr << "Could not read file " << file << '\n';
			inputName[0]='\0';
		}
		pos=BUFFERSIZE;
		n_chars=0;
		current_line=current_pos=1;
	}
	~ReadFileObject() {
		in.close();
	}
};

class ParserObject {
public:

	// This is (context switchable) info on the file being read.

	ReadFileObject *current_file;
	DLList<ReadFileObject> file_Stack;
	void push_buffer(const char* file, DLList<SFNode> *list) {
		file_Stack.add(current_file);
		current_file=new ReadFileObject(file);
		if (!current_file) {
			std::cerr << "Ran out of memory in push_buffer()\n";
            abort();
		}
		node_Stack.add(current_Nodes);
		current_Nodes=list;
	}
	void pop_buffer(void) {
		delete current_file;
		current_file=file_Stack.current();
        file_Stack.remove();
		current_Nodes=node_Stack.current();
        node_Stack.remove();
	}

	// This is an array of pointers to functions used to parse types.
	fieldType* (*parse_type[21])(ParserObject*);

	//
	// This is an array of names used for convenience later if I want to
	// output the name of a type.
	//
	char* type_name[21];

	int isVerbose;
	void verbose(void) { isVerbose=TRUE; }
	int parse(const char* name);

	//
	// These hold the context-sensitive definitions of PROTO's and DEF's.
	// The root* lists contain those
	// items available in the root VRML scene. These, however, are context
	// dependent, as each PROTO statement has its own independent definitions
	// of local PROTO's and DEF's. So, when the parser descends into PROTO's
	// (and their children),
	// the current_* pointers are updated so that PROTO's and DEF's have
	// the correct scope.
	//

    DLList<protoStatement> root_PROTOs, *local_PROTOs;
	DLList< DLList<protoStatement> > PROTO_Stack;

    // find named PROTO given current local PROTO's and those from outer contexts
    protoStatement* FindPROTO(const char *name, ReadFileObject *r);
    
	DLList< DLList<interfaceDeclaration> > arg_Stack;
	DLList<SFNode> root_Nodes, *current_Nodes;
	DLList< DLList<SFNode> > node_Stack;
	DLList<nodeStatement> root_DEFs, *current_DEFs;
	DLList< DLList<nodeStatement> > DEF_Stack;
	int in_PROTO, in_PROTO_declarations;

	void push_context(protoStatement* pr) {
        PROTO_Stack.add(local_PROTOs);
        local_PROTOs=&(pr->local_PROTOs);
		arg_Stack.add(&(pr->interfaceDeclarations));
		node_Stack.add(current_Nodes);
		current_Nodes=&(pr->local_Nodes);
		DEF_Stack.add(current_DEFs);
		current_DEFs=&(pr->local_DEFs);
		in_PROTO++;
	}
	void pop_context(void) {
        local_PROTOs=PROTO_Stack.current();
        PROTO_Stack.remove();
        arg_Stack.remove();
		current_Nodes=node_Stack.current();
        node_Stack.remove();
		current_DEFs=DEF_Stack.current();
        DEF_Stack.remove();
		in_PROTO--;
	}

	ParserObject() {
		isVerbose=FALSE;
        local_PROTOs=&root_PROTOs;
		current_Nodes=&root_Nodes;
		current_DEFs=&root_DEFs;
		in_PROTO=in_PROTO_declarations=FALSE;
        parse_type[ERROR]=NULL;                  type_name[ERROR]="(no type!)";
		parse_type[SFBOOL]=parse_SFBool;	     type_name[SFBOOL]="SFBool";
		parse_type[SFCOLOR]=parse_SFColor;	     type_name[SFCOLOR]="SFColor";
		parse_type[MFCOLOR]=parse_MFColor;	     type_name[MFCOLOR]="MFColor";
		parse_type[SFFLOAT]=parse_SFFloat;	     type_name[SFFLOAT]="SFFloat";
		parse_type[MFFLOAT]=parse_MFFloat;	     type_name[MFFLOAT]="MFFloat";
		parse_type[SFIMAGE]=parse_SFImage;	     type_name[SFIMAGE]="SFImage";
		parse_type[SFINT32]=parse_SFInt32;       type_name[SFINT32]="SFInt32";
		parse_type[MFINT32]=parse_MFInt32;	     type_name[MFINT32]="MFInt32";
		parse_type[SFNODE]=parse_SFNode;	     type_name[SFNODE]="SFNode";
		parse_type[MFNODE]=parse_MFNode;	     type_name[MFNODE]="MFNode";
		parse_type[SFROTATION]=parse_SFRotation; type_name[SFROTATION]="SFRotation";
		parse_type[MFROTATION]=parse_MFRotation; type_name[MFROTATION]="MFRotation";
		parse_type[SFSTRING]=parse_SFString;     type_name[SFSTRING]="SFString";
		parse_type[MFSTRING]=parse_MFString;     type_name[MFSTRING]="MFString";
		parse_type[SFTIME]=parse_SFTime;	     type_name[SFTIME]="SFTime";
		parse_type[MFTIME]=parse_MFTime;	     type_name[MFTIME]="MFTime";
		parse_type[SFVEC2F]=parse_SFVec2f;	     type_name[SFVEC2F]="SFVec2f";
		parse_type[MFVEC2F]=parse_MFVec2f;	     type_name[MFVEC2F]="MFVec2f";
		parse_type[SFVEC3F]=parse_SFVec3f;       type_name[SFVEC3F]="SFVec3f";
		parse_type[MFVEC3F]=parse_MFVec3f;	     type_name[MFVEC3F]="MFVec3f";
	}

#define DELETE_ALL_IN_LIST_OF_LISTS(LIST) \
    for (i=0; i<(LIST).n_items(); i++) { \
        (LIST).current()->empty_and_delete_all(); \
        (LIST).next(); \
    } \
    (LIST).empty_and_delete_all();

    ~ParserObject() {
        int i;
        fieldType::FreeAllocatedNodes();
        file_Stack.empty_and_delete_all();
        root_PROTOs.empty_and_delete_all();
        DELETE_ALL_IN_LIST_OF_LISTS(PROTO_Stack);
        DELETE_ALL_IN_LIST_OF_LISTS(arg_Stack);
        node_Stack.empty_and_delete_all();
        root_DEFs.empty_and_delete_all();
        DELETE_ALL_IN_LIST_OF_LISTS(DEF_Stack);
    }
};

inline int is_line_terminator(char c) // CR, LF
{
	if (c==0x0a || c==0x0d)
		return TRUE;
	else
		return FALSE;
}

inline int is_ws(char c) // CR, LF, space, tab, comma
{
	if (c==0x0a || c==0x0d || c==0x20 || c==0x09 || c==0x2c)
		return TRUE;
	else
		return FALSE;
}

inline int is_enclosure(char c) // { } [ ] "
{
	if (c=='{' || c=='}' || c=='[' || c==']' || c=='"')
		return TRUE;
	else
		return FALSE;
}

extern int is_valid_id(char *word);

//
// A buffered read of a character; '\0' is returned if EOF
//

inline char ReadFileObject::get_ch(void)
{
	if (pos>=BUFFERSIZE) {
		in.read(buffer,BUFFERSIZE);
		n_chars=in.gcount();
		pos=0;
	}
	if (buffer[pos]==0x0a || buffer[pos]==0x0d)
		current_pos=0;
	else
		current_pos++;
	if (buffer[pos]==0x0a) current_line++;
	if (pos<n_chars)
		return buffer[pos++];
	else
		return '\0';
}

// Peek at the next character, but do not retrieve it.

inline char ReadFileObject::peek_ch(void)
{
	if (pos>=BUFFERSIZE) {
		int ch=in.peek();
		if (ch!=EOF)
			return (char)ch;
		else
			return '\0';
	}
	if (pos<n_chars)
		return buffer[pos];
	else
		return '\0';
}

inline protoStatement* find_PROTO(DLList<protoStatement> *list, const char* id)
{
	for (int i=0; i<list->n_items(); i++) {
		if (strcmp(id,list->current()->nodeTypeId)==0)
			return list->current();
		list->next();
	}
	return NULL;
}

int parse_file(ParserObject*);
int skip_route(ParserObject*);

#endif // __vrml2pov_parser_h_
