/*
* pov.cpp
*
* $Id: pov.cpp,v 1.6 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* The functions that read the node tree generated by the parser, outputting
* POV objects as appropriate.
*/

#include <fstream>

#define __vrml2pov_pov_cc_
#include "pov.h"
#include "parser.h"
#include "node.h"
#include "fieldtyp.h"
#include "povout.h"

using namespace std;

int verbose=FALSE;
ofstream out;
DLList<View_transform> view_transform_stack;
int n_lights=0;
double headlight=1.0, line_radius=0.05, point_radius=0.05;
int do_backface_culling=FALSE;
F_TYPE* culling_viewpoint=NULL;

nodeVariable* find_field(SFNode* node, const char* id)
{
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		if (strcmp(node->nodeBody.current()->fieldId,id)==0)
			return node->nodeBody.current();
		node->nodeBody.next();
	}
	cerr << "Internal error: " << node->nodeTypeId << " missing field "
		<< id << '\n';
    abort();
	return NULL; // compiler complains about no return type...
}

void do_POV_output(ParserObject* p, const char* infile, const char* outfile)
{
	DLList<SFNode>& root=p->root_Nodes; // for convenience

	if (p->isVerbose) verbose=TRUE;
	finish[0]='\0';

	// if backface culling is to be done, the first node must be Viewpoint

	root.start();
	if (do_backface_culling) {
		if (strcmp(root.current()->nodeTypeId,"Viewpoint")!=0) {
			cerr <<
				"If backface culling is to be done, the first node must be Viewpoint.\n";
			return;
		}
		culling_viewpoint=FIND(root.current(),position,SFVec3f)->value;
	}


	// check and open output

	out.open(outfile,ios::out|ios::trunc);
	if (out.fail()) {
		cerr << "Can't write to file " << outfile << '\n';
		return;
	}

	// output a header; include WorldInfo stuff if present at root level

	out << "/*\n\n  Created from \"" << infile << "\" by\n" << VERSION;

	int i, j;
	for (i=0; i<root.n_items(); i++) {
		if (strcmp(root.current()->nodeTypeId,"WorldInfo")==0) {
			out << "  Title: " << FIND(root.current(),title,SFString)->value << '\n';
			out << "  other info:\n";
			for (j=0; j<FIND(root.current(),info,MFString)->n_items; j++) {
				out << "    " << FIND(root.current(),info,MFString)->value[j]
					<< '\n';
			}
			break;
		}
		root.next();
	}
	out << "\n*/\n\n";

	// set default precision for floating point values

	out.precision(6);

	//
	// Initialize the current transformation. This variable keeps track of the
	// local coordinate system as Transform nodes are encountered, so that their
	// children are affected appropriately.
	//

	current_transformation=new Matrix();
	if (!current_transformation) MEMOUT;

	// Descend the node tree and deal with each node as appropriate

	//  out << "#declare VRML_scene = union {\n";
	out << "union {\n";
	indent=ii;
	root.start();
	for (i=0; i<root.n_items(); i++) {
		do_node(root.current());
		root.next();
	}

	//
	// output the camera and transform the scene. In theory, one should be
	// able to apply the camera transformation (as determined directly by the
	// Viewpoint node) to the POV camera. However, POV is finicky about applying
	// matrices to the camera, so it proved easier to leave the camera at the
	// origin and apply the inverse transformation to the rest of the scene...
	//

	if (view_transform_stack.n_items()) {
		Matrix inverse;
		view_transform_stack.start();
		for (i=0; i<view_transform_stack.n_items(); i++) {
			invert_into(&(view_transform_stack.current()->m),&inverse);
			if (i==0) {
				out <<
					"\n// This matrix is for viewing only; remove it to leave\n"
					<<
					"// objects at their world coordinates as specified in the VRML file\n"
					<< INDENT << &inverse
					<< "\n\n}\n\n"
					<<
					"// Note that the camera used here defines a right-handed universe\n"
					<<
					"// (to change view, substitute above matrix, and camera angle)\n\n";
			}
			out << "// for Viewpoint \""
				<< view_transform_stack.current()->description << "\"\n"
				<< "// to view = " << &inverse << '\n'
				<< "// angle = "
				<< view_transform_stack.current()->angle << "\n\n";
			view_transform_stack.next();
		}
		if (i>1) invert_into(&(view_transform_stack.first()->m),&inverse);
		out << "camera {              // for "
			<< view_transform_stack.first()->description << '\n'
			<< "  location <0,0,0>\n"
			<< "  look_at <0,0,-1>\n"
            << "  direction <0,0,-1>\n"
			<< "  sky <0,1,0>\n"
			<< "  up <0,1,0>\n"
			<< "  right <4/3,0,0>\n"
			<< "  angle " << view_transform_stack.first()->angle << '\n'
			<< "}\n\n";
		if (headlight>0.0)
			out << "light_source { <0,0,0> rgb " << headlight << "*<1,1,1> }\n";
	} else {
		out << "}\n\n" <<
			"// Note that the camera used here defines a right-handed universe\n\n"
			<< "camera {\n"
			<< INDENT << "location <0,0,10>\n"
			<< INDENT << "look_at <0,0,0>\n"
            << INDENT << "direction <0,0,-1>\n"
			<< INDENT << "sky <0,1,0>\n"
			<< INDENT << "up <0,1,0>\n"
			<< INDENT << "right <4/3,0,0>\n"
			<< INDENT << "angle 45\n"
			<< "}\n\n";
		if (headlight>0.0)
			out << "light_source { <0,0,10> rgb " << headlight << "*<1,1,1> }\n";
	}

	indent=0;
	out.close();

    // cleanup
    view_transform_stack.empty_and_delete_all();;
    delete current_transformation;

	return;
}


//
// First determine what type of node we're looking at. Then call the
// appropriate POV output function for that node, if any. Basically, this
// routine is just a link to functions that actually create output.
//

void do_node(SFNode* node)
{
	if (strcmp(node->nodeTypeId,"Appearance")==0) do_Appearance(node);
	else if (strcmp(node->nodeTypeId,"Box")==0) do_Box(node);
	else if (strcmp(node->nodeTypeId,"Collision")==0) do_Collision(node);
	else if (strcmp(node->nodeTypeId,"Cone")==0) do_Cone(node);
	else if (strcmp(node->nodeTypeId,"Cylinder")==0) do_Cylinder(node);
	else if (strcmp(node->nodeTypeId,"Extrusion")==0) do_Extrusion(node);
	else if (strcmp(node->nodeTypeId,"Group")==0) do_Group(node);
	else if (strcmp(node->nodeTypeId,"IndexedFaceSet")==0) do_IndexedFaceSet(node);
	else if (strcmp(node->nodeTypeId,"IndexedLineSet")==0) do_IndexedLineSet(node);
	else if (strcmp(node->nodeTypeId,"Inline")==0) do_Inline(node);
	else if (strcmp(node->nodeTypeId,"Material")==0) do_Material(node);
	else if (strcmp(node->nodeTypeId,"PointSet")==0) do_PointSet(node);
	else if (strcmp(node->nodeTypeId,"Shape")==0) do_Shape(node);
	else if (strcmp(node->nodeTypeId,"Sphere")==0) do_Sphere(node);
	else if (strcmp(node->nodeTypeId,"Transform")==0) do_Transform(node);
	else if (strcmp(node->nodeTypeId,"Viewpoint")==0) do_Viewpoint(node);
	else if (strcmp(node->nodeTypeId,"WorldInfo")==0) return;
	else {
		if (verbose) cout << "No POV output for node " << node->nodeTypeId
			<< "; contents ignored.\n";
	}
	return;
}
