/*
* povout.cpp
*
* $Id: povout.cpp,v 1.5 2003/02/18 23:23:53 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* POV output routines for comparatively simple VRML objects; more complex ones
* may be in separate modules.
*/

#define __vrml2pov_povout_cc_

#include <strstream>

#include "matrix.h"
#include "pov.h"

using namespace std;

Matrix* current_transformation;
DLList<Matrix> transformation_Stack;
int indent=0, ii=2;
int use_transparency=FALSE;
F_TYPE transparency;
char finish[FINISH_SIZE];
SFColor* emissiveColor;

#include "povout.h"

void do_Transform(SFNode* node)
{
	static Matrix T, C, R, SR, S, nSR, nC;
	set_translation_matrix(&T,FIND(node,translation,SFVec3f));
	set_translation_matrix(&C,FIND(node,center,SFVec3f));
	set_rotation_matrix(&R,FIND(node,rotation,SFRotation));
	set_rotation_matrix(&SR,FIND(node,scaleOrientation,SFRotation));
	set_scale_matrix(&S,FIND(node,scale,SFVec3f));
	set_rotation_matrix(&nSR,FIND(node,scaleOrientation,SFRotation),-1);
	set_translation_matrix(&nC,FIND(node,center,SFVec3f),-1);
	
	Matrix* new_transformation=new Matrix();
	if (!new_transformation) MEMOUT;
	static Matrix tmp1, tmp2;
	compose_into(&nSR,&nC,&tmp1);
	compose_into(&S,&tmp1,&tmp2);
	compose_into(&SR,&tmp2,&tmp1);
	compose_into(&R,&tmp1,&tmp2);
	compose_into(&C,&tmp2,&tmp1);
	compose_into(&T,&tmp1,&tmp2);
	compose_into(current_transformation,&tmp2,new_transformation);
	
	MFNode* children=FIND(node,children,MFNode);
	
	//
	// descend into children with new local coordinate transformation; restore
	// old one when done.
	//
	
	push_transformation(new_transformation);
	for (int i=0; i<children->n_items; i++)
		do_node(children->value[i]);
	pop_transformation(); // also deletes new_transformation
	
	return;
}

//
// This function is *not* used to output the matrix verbatim, but to
// create a matrix <> transformation object as POV understands it.
//
ostream& operator<<(ostream& s, Matrix* m)
{
	return s << "matrix <"
		<< m->m[0] << ',' << m->m[4] << ',' << m->m[8] << ','
		<< m->m[1] << ',' << m->m[5] << ',' << m->m[9] << ','
		<< m->m[2] << ',' << m->m[6] << ',' << m->m[10] << ','
		<< m->m[3] << ',' << m->m[7] << ',' << m->m[11] << '>';
}

//
// translates an Appearance material into a POV texture. However, no
// texturing is currently possible.
//

void do_Appearance(SFNode* node)
{
	SFNode* material=FIND(node,material,SFNode);
	if (material) do_Material(material);
	if (verbose &&
		(FIND(node,texture,SFNode) || FIND(node,textureTransform,SFNode)))
		cerr << "Warning: texturing (via Appearance) not currently implemented.\n";
	return;
}

// output Box

void do_Box(SFNode* node)
{
	F_TYPE *corner=FIND(node,size,SFVec3f)->value;
	F_TYPE x=corner[0]/2.0, y=corner[1]/2.0, z=corner[2]/2.0;
	out << INDENT << "box { <" << x << ',' << y << ',' << z << ">, <"
		<< -x << ',' << -y << ',' << -z << "> "
		<< current_transformation;
	if (finish[0]) out << ' ' << finish;
	out << " }\n";
	return;
}

// for Collision, just descend into its children

void do_Collision(SFNode* node)
{
	MFNode* children=FIND(node,children,MFNode);
	for (int i=0; i<children->n_items; i++)
		do_node(children->value[i]);
	return;
}

// output Cone

void do_Cone(SFNode* node)
{
	F_TYPE height=FIND(node,height,SFFloat)->value;
	int side=FIND(node,side,SFBool)->value;
	int bottom=FIND(node,bottom,SFBool)->value;
	if (side) {
		out << INDENT << "cone { <0," << height/2.0 << ",0>, 0, <0,"
			<< -height/2.0 << ",0>, " << FIND(node,bottomRadius,SFFloat)->value;
		if (!bottom) out << " open";
	} else if (bottom) {
		out << INDENT << "disc { <0," << -height/2.0 << ",0>, <0,1,0>, "
			<< FIND(node,bottomRadius,SFFloat)->value;
	}
	if (side || bottom) {
		out << ' ' << current_transformation;
		if (finish[0]) out << ' ' << finish;
		out << " }\n";
	}
	return;
}

// output Cylinder

void do_Cylinder(SFNode* node)
{
	F_TYPE height=FIND(node,height,SFFloat)->value;
	F_TYPE radius=FIND(node,radius,SFFloat)->value;
	int top=FIND(node,top,SFBool)->value;
	int bottom=FIND(node,bottom,SFBool)->value;
	int side=FIND(node,side,SFBool)->value;
	if (!side && !top && !bottom) return;
	int needUnion=( (side && ((!top && bottom) || (top && !bottom))) ||
		(!side && (top && bottom)) );
	if (needUnion) {
		out << INDENT << "union {\n";
		indent+=ii;
	}
	if (side) {
		out << INDENT << "cylinder { <0," << height/2.0 << ",0>, <0,"
			<< -height/2.0 << ",0>, " << radius;
		if (!top && !bottom)
			out << " open";
		else if (top && !bottom)
			out << " open }\n" << INDENT << "disc { <0," << height/2.0
			<< ",0>, <0,1,0>, " << radius;
		else if (!top && bottom)
			out << " open }\n" << INDENT << "disc { <0," << -height/2.0
			<< ",0>, <0,-1,0>, " << radius;
	} else {
		if (top)
			out << INDENT << "disc { <0," << height/2.0
			<< ",0>, <0,1,0>, " << radius;
		if (bottom) {
			if (top) out << " }\n";
			out << INDENT << "disc { <0," << -height/2.0
				<< ",0>, <0,-1,0>, " << radius;
		}
	}
	if (needUnion) {
		out << " }\n" << INDENT << current_transformation << '\n';
		if (finish[0]) out << INDENT << finish << '\n';
		indent-=ii;
		out << INDENT << "}\n";
	} else {
		out << ' ' << current_transformation;
		if (finish[0]) out << ' ' << finish;
		out << " }\n";
	}
	return;
}

// for Group, just descend into its children

void do_Group(SFNode* node)
{
	MFNode* children=FIND(node,children,MFNode);
	for (int i=0; i<children->n_items; i++)
		do_node(children->value[i]);
	return;
}

//
// for Inline, just descend into the "children" created by the parser. Note
// that this is not a valid construct according to the specification, but
// it's just a convenient way to hold the nodes created by an Inline file
// in the correct place.
//

void do_Inline(SFNode* node)
{
	MFNode* children=FIND(node,children,MFNode);
	for (int i=0; i<children->n_items; i++)
		do_node(children->value[i]);
	return;
}

//
// Try to translate a Material into a POV texture. This is crude at
// the moment... POV's lighting model is different from VRML's, so the
// shading will likely be slightly different from what a VRML browser might
// show.
//

void do_Material(SFNode* node)
{
	F_TYPE t=FIND(node,transparency,SFFloat)->value;
	Vector color=FIND(node,diffuseColor,SFColor),
		spec=FIND(node,specularColor,SFColor);
	
	ostrstream fstr(finish,FINISH_SIZE);
	
	fstr << "pigment { rgb";
	if (t) {
		fstr << "t <" << color.v[0] << ',' << color.v[1] << ','
			<< color.v[2] << ',' << t << "> } ";
		transparency=t;
		use_transparency=TRUE;
	} else
		fstr << ' ' << color << " } ";
	
	//  fstr << "finish { diffuse 0.8 ambient "
	//	<< FIND(node,ambientIntensity,SFFloat)->value
	F_TYPE ambient=FIND(node,ambientIntensity,SFFloat)->value;
	ambient-=0.3;
	if (ambient<0.0) ambient=0.0;
	F_TYPE s=sqrt((spec.v[0])*(spec.v[0])+
		(spec.v[1])*(spec.v[1])+
		(spec.v[2])*(spec.v[2]))/3.0;
	fstr << "finish { diffuse 0.9 ambient " << ambient
		<< " specular " << s;
	if (s>0.0001)
		fstr << " roughness " << 0.2505-(FIND(node,shininess,SFFloat)->value)/4.0;
	fstr << " }" << '\0';
	emissiveColor=FIND(node,emissiveColor,SFColor);
	if (verbose &&
		(emissiveColor->value[0]!=0.0 || emissiveColor->value[1]!=0.0 ||
		emissiveColor->value[2]!=0.0))
		cout << "Warning: emissive color not currently supported,\n"
		<< "  except to color IndexedLineSet and PointSet.\n";
	return;
}

//
// A Shape is just an object attached to an Appearance, so write a POV
// object{} statement with the object and a texture{}.
//

void do_Shape(SFNode* node)
{
	// try to ensure that only values from current Material are used
	
	finish[0]='\0';
	use_transparency=FALSE;
	emissiveColor=NULL;
	
	SFNode* texture=FIND(node,appearance,SFNode);
	if (!texture) {
		if (verbose) cerr << "Warning: NULL appearance in Shape node\n";
	} else {
		if (strcmp(texture->nodeTypeId,"Appearance")) {
			cerr << "Internal error: somehow got a non-Appearance node in the\n"
				<< "appearance field of Shape\n";
			return;
		}
	}
	
	SFNode* object=FIND(node,geometry,SFNode);
	if (!object) {
        if (verbose) cerr << "Warning: Shape node has no geometry!\n";
		return;
	}
	if (strcmp(object->nodeTypeId,"Cylinder")!=0 &&
		strcmp(object->nodeTypeId,"Box")!=0 &&
		strcmp(object->nodeTypeId,"Cone")!=0 &&
		strcmp(object->nodeTypeId,"Extrusion")!=0 &&
		strcmp(object->nodeTypeId,"IndexedFaceSet")!=0 &&
		strcmp(object->nodeTypeId,"IndexedLineSet")!=0 &&
		strcmp(object->nodeTypeId,"PointSet")!=0 &&
		strcmp(object->nodeTypeId,"Sphere")!=0) {
		if (verbose)
			cerr << object->nodeTypeId
			<< " is either not a Shape geometry, or not yet implemented...\n";
		return;
	}
	
	if (texture)
		do_Appearance(texture); // writes a POV texture{} to global string "finish"
	do_node(object); // writes out the object (the node itself uses the finish)
	
	// clear these so they're not used by something else
	
	finish[0]='\0';
	use_transparency=FALSE;
	emissiveColor=NULL;
	
	return;
}

// output Sphere

void do_Sphere(SFNode* node)
{
	out << INDENT << "sphere { <0,0,0>, " << FIND(node,radius,SFFloat)->value
		<< ' ' << current_transformation;
	if (finish[0]) out << ' ' << finish;
	out << " }\n";
	return;
}

//
// For now, the first Viewpoint found is the one used; however, for
// convenience, all other Viewpoints are output inside comment markers
// so that they can be easily accessed by commenting/uncommenting in the
// POV file.
//

void do_Viewpoint(SFNode* node)
{
	static Matrix pos, orient, tmp;
	View_transform* view=new View_transform();
	if (!view) MEMOUT;
	
	set_translation_matrix(&pos,FIND(node,position,SFVec3f));
	set_rotation_matrix(&orient,FIND(node,orientation,SFRotation));
	//  compose_into(&orient,&pos,&tmp);
	//  compose_into(current_transformation,&tmp,&(view->m));
	compose_into(&pos,&orient,&tmp);
	compose_into(current_transformation,&tmp,&(view->m));
    view->angle=(180.0/acos(-1.0))*FIND(node,fieldOfView,SFFloat)->value;
	char *d=FIND(node,description,SFString)->value;
	view->description=new char[strlen(d)+1];
	if (!view->description) MEMOUT;
	strcpy(view->description,d);
	
	view_transform_stack.add(view);
	return;
}
