/*
* readtype.cpp
*
* $Id: readtype.cpp,v 1.4 2003/02/18 23:23:54 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* ParserObject functions to parse each and store each fieldType. In each
* case a pointer to a new object of that type is returned, or NULL if
* a parsing error occurs. Of course, the reading of MF* types is complicated
* by the fact that when the parser begins reading the multiple values, it
* doesn't know how many there will be - and thus how much memory to
* allocate. I've gotten around this problem by storing individual values
* in a temporary buffer list, then creating an MF* of just the right size
* and moving values from the buffer.
*/

#include "parser.h"

using namespace std;

//
// Returns an identifier for the fieldType string.
//
unsigned char get_type(const char* s)
{
	if (strcmp(s,"SFBool")==0) return SFBOOL;
	else if (strcmp(s,"SFColor")==0) return SFCOLOR;
	else if (strcmp(s,"MFColor")==0) return MFCOLOR;
	else if (strcmp(s,"SFFloat")==0) return SFFLOAT;
	else if (strcmp(s,"MFFloat")==0) return MFFLOAT;
	else if (strcmp(s,"SFImage")==0) return SFIMAGE;
	else if (strcmp(s,"SFInt32")==0) return SFINT32;
	else if (strcmp(s,"MFInt32")==0) return MFINT32;
	else if (strcmp(s,"SFNode")==0) return SFNODE;
	else if (strcmp(s,"MFNode")==0) return MFNODE;
	else if (strcmp(s,"SFRotation")==0) return SFROTATION;
	else if (strcmp(s,"MFRotation")==0) return MFROTATION;
	else if (strcmp(s,"SFString")==0) return SFSTRING;
	else if (strcmp(s,"MFString")==0) return MFSTRING;
	else if (strcmp(s,"SFTime")==0) return SFTIME;
	else if (strcmp(s,"MFTime")==0) return MFTIME;
	else if (strcmp(s,"SFVec2f")==0) return SFVEC2F;
	else if (strcmp(s,"MFVec2f")==0) return MFVEC2F;
	else if (strcmp(s,"SFVec3f")==0) return SFVEC3F;
	else if (strcmp(s,"MFVec3f")==0) return MFVEC3F;
	else return ERROR;
}

//
// These routines parse and create storage for each fieldType. The
// current_word is assumed in each case to be the first word of the
// field to be parsed, and the routines leave the current_word on
// the last word of the field.
//

#define GET_FLOAT(VAR) do { \
    if ((r.current_word[0]>='0' && r.current_word[0]<='9') || \
        r.current_word[0]=='+' || r.current_word[0]=='-' || \
        r.current_word[0]=='.') \
        VAR=atof(r.current_word); \
    else \
        return (fieldType*)r.parse_error("floating point value",r.current_word); \
} while (0)

#define GET_INT(VAR) do { \
    if ((r.current_word[0]>='0' && r.current_word[0]<='9') || \
        r.current_word[0]=='+' || r.current_word[0]=='-') \
        VAR=atoi(r.current_word); \
    else \
        return (fieldType*)r.parse_error("integer value",r.current_word); \
} while (0)

fieldType* parse_SFBool(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFBool* obj=new SFBool();
	if (!obj) OUTMEM;
	if (strcmp(r.current_word,"TRUE")==0) obj->value=TRUE;
	else if (strcmp(r.current_word,"FALSE")==0) obj->value=FALSE;
	else return (fieldType*)r.parse_error("SFBool",r.current_word);
	return obj;
}

SFBool* copy_SFBool(SFBool* n, SFBool* old)
{
	n->value=old->value;
	return n;
}

fieldType* parse_SFColor(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFColor* obj=new SFColor();
	if (!obj) OUTMEM;
	GET_FLOAT(obj->value[0]);
	if (obj->value[0]<0.0 || obj->value[0]>1.0)  // warning only
		r.parse_error("red value from 0.0 to 1.0",r.current_word);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("green value",NULL);
	GET_FLOAT(obj->value[1]);
	if (obj->value[1]<0.0 || obj->value[1]>1.0)  // warning only
		r.parse_error("green value from 0.0 to 1.0",r.current_word);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("blue value",NULL);
	GET_FLOAT(obj->value[2]);
	if (obj->value[2]<0.0 || obj->value[2]>1.0)  // warning only
		r.parse_error("blue value from 0.0 to 1.0",r.current_word);
	return obj;
}

SFColor* copy_SFColor(SFColor* n, SFColor* old)
{
	for (int i=0; i<3; i++) n->value[i]=old->value[i];
	return n;
}

fieldType* parse_MFColor(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	F_TYPE *color;
	MFColor* mfc;
	static DLList<F_TYPE> buffer;
	unsigned char multiple=FALSE;
	if (strcmp(r.current_word,"[")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("red value",NULL);
		multiple=TRUE;
	}
	while (strcmp(r.current_word,"]")!=0) {
		color=new F_TYPE[3];
		if (!color) OUTMEM;
		GET_FLOAT(color[0]);
		if (color[0]<0.0 || color[0]>1.0)  // warning only
			r.parse_error("red value from 0.0 to 1.0",r.current_word);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("green value",NULL);
		GET_FLOAT(color[1]);
		if (color[1]<0.0 || color[1]>1.0)  // warning only
			r.parse_error("green value from 0.0 to 1.0",r.current_word);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("blue value",NULL);
		GET_FLOAT(color[2]);
		if (color[2]<0.0 || color[2]>1.0)  // warning only
			r.parse_error("blue value from 0.0 to 1.0",r.current_word);
		buffer.add(color);
		if (multiple) {
			if (!r.get_next_word())
				return (fieldType*)r.parse_error("red value or ]",NULL);
		} else
			break;
	}
	mfc=new MFColor(buffer.n_items());
	if (!mfc) OUTMEM;
	buffer.start();
	for (int i=0; i<buffer.n_items(); i++) {
		mfc->value[3*i]=buffer.current()[0];
		mfc->value[3*i+1]=buffer.current()[1];
		mfc->value[3*i+2]=buffer.current()[2];
		buffer.next();
	}
    buffer.empty_and_delete_all();
	return mfc;
}

MFColor* copy_MFColor(MFColor* n, MFColor* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new F_TYPE[old->n_items*3];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items*3; i++) n->value[i]=old->value[i];
	}
	return n;
}

fieldType* parse_SFFloat(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFFloat* obj=new SFFloat();
	if (!obj) OUTMEM;
	GET_FLOAT(obj->value);
	return obj;
}

SFFloat* copy_SFFloat(SFFloat* n, SFFloat* old)
{
	n->value=old->value;
	return n;
}

fieldType* parse_MFFloat(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	F_TYPE *val;
	MFFloat* mff;
	static DLList<F_TYPE> buffer;
	unsigned char multiple=FALSE;
	if (strcmp(r.current_word,"[")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("floating point value",NULL);
		multiple=TRUE;
	}
	while (strcmp(r.current_word,"]")!=0) {
		val=new F_TYPE[1];
		if (!val) OUTMEM;
		GET_FLOAT(*val);
		buffer.add(val);
		if (multiple) {
			if (!r.get_next_word())
				return (fieldType*)r.parse_error("floating point value or ]",NULL);
		} else
			break;
	}
	mff=new MFFloat(buffer.n_items());
	if (!mff) OUTMEM;
	buffer.start();
	for (int i=0; i<buffer.n_items(); i++) {
		mff->value[i]=*(buffer.current());
		buffer.next();
	}
    buffer.empty_and_delete_all();
	return mff;
}

MFFloat* copy_MFFloat(MFFloat* n, MFFloat* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new F_TYPE[old->n_items];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items; i++) {
			n->value[i]=old->value[i];
		}
	}
	return n;
}

//
// Pixels are stored as integer values; so this only works if the #bytes
// in an integer is <= # byte-components of each pixel...
//

fieldType* parse_SFImage(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	int w, h, n, i, l;

	GET_INT(w);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("height value",NULL);
	GET_INT(h);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("#components value",NULL);
	GET_INT(n);
	if (n>((int)sizeof(SFIMAGE_INT_TYPE))) {
		cerr << "Implementation error: int "
            << "values on this system are too small\n"
			<< "to hold " << n
			<< "-component pixels - make SFIMAGE_INT_TYPE\n"
			<< "in fieldtyp.h a bigger integer type.\n";
		return NULL;
	}
	SFImage* image=new SFImage(w,h,n);
	if (!image) OUTMEM;
	l=w*h;
	for (i=0; i<l; i++) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("pixel value",NULL);
		GET_INT(image->pixel_data[i]);
	}
	return image;
}

SFImage* copy_SFImage(SFImage* n, SFImage* old)
{
	n->width=old->width;
	n->height=old->height;
	n->n_components=old->n_components;
	int l=n->width*n->height;
	n->pixel_data=new SFIMAGE_INT_TYPE[l];
	if (!n->pixel_data) OUTMEM;
	for (int i=0; i<l; i++) n->pixel_data[i]=old->pixel_data[i];
	return n;
}

fieldType* parse_SFInt32(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFInt32* obj=new SFInt32();
	if (!obj) OUTMEM;
	GET_INT(obj->value);
	return obj;
}

SFInt32* copy_SFInt32(SFInt32* n, SFInt32* old)
{
	n->value=old->value;
	return n;
}

fieldType* parse_MFInt32(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	I_TYPE *val;
	MFInt32* mfi;
	static DLList<I_TYPE> buffer;
	unsigned char multiple=FALSE;
	if (strcmp(r.current_word,"[")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("integer value",NULL);
		multiple=TRUE;
	}
	while (strcmp(r.current_word,"]")!=0) {
		val=new I_TYPE[1];
		if (!val) OUTMEM;
		GET_INT(*val);
		buffer.add(val);
		if (multiple) {
			if (!r.get_next_word())
				return (fieldType*)r.parse_error("integer value or ]",NULL);
		} else
			break;
	}
	mfi=new MFInt32(buffer.n_items());
	if (!mfi) OUTMEM;
	buffer.start();
	for (int i=0; i<buffer.n_items(); i++) {
		mfi->value[i]=*(buffer.current());
		buffer.next();
	}
    buffer.empty_and_delete_all();
	return mfi;
}

MFInt32* copy_MFInt32(MFInt32* n, MFInt32* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new I_TYPE[old->n_items];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items; i++) {
			n->value[i]=old->value[i];
		}
	}
	return n;
}

// (SFNodes and MFNodes parsed in node.cpp)

fieldType* parse_SFRotation(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFRotation* obj=new SFRotation();
	if (!obj) OUTMEM;
	GET_FLOAT(obj->value[0]);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("y value",NULL);
	GET_FLOAT(obj->value[1]);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("z value",NULL);
	GET_FLOAT(obj->value[2]);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("rotation value",NULL);
	GET_FLOAT(obj->value[3]);
	return obj;
}

SFRotation* copy_SFRotation(SFRotation* n, SFRotation* old)
{
	for (int i=0; i<4; i++) n->value[i]=old->value[i];
	return n;
}

fieldType* parse_MFRotation(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	F_TYPE *sfr;
	MFRotation* mfr;
	static DLList<F_TYPE> buffer;
	unsigned char multiple=FALSE;
	if (strcmp(r.current_word,"[")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("x value",NULL);
		multiple=TRUE;
	}
	while (strcmp(r.current_word,"]")!=0) {
		sfr=new F_TYPE[4];
		if (!sfr) OUTMEM;
		GET_FLOAT(sfr[0]);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("y value",NULL);
		GET_FLOAT(sfr[1]);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("z value",NULL);
		GET_FLOAT(sfr[2]);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("rotation value",NULL);
		GET_FLOAT(sfr[3]);
		buffer.add(sfr);
		if (multiple) {
			if (!r.get_next_word())
				return (fieldType*)r.parse_error("x value or ]",NULL);
		} else
			break;
	}
	mfr=new MFRotation(buffer.n_items());
	if (!mfr) OUTMEM;
	buffer.start();
	for (int i=0; i<buffer.n_items(); i++) {
		mfr->value[4*i]=buffer.current()[0];
		mfr->value[4*i+1]=buffer.current()[1];
		mfr->value[4*i+2]=buffer.current()[2];
		mfr->value[4*i+3]=buffer.current()[3];
		buffer.next();
	}
    buffer.empty_and_delete_all();
	return mfr;
}

MFRotation* copy_MFRotation(MFRotation* n, MFRotation* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new F_TYPE[old->n_items*4];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items*4; i++) n->value[i]=old->value[i];
	}
	return n;
}

fieldType* parse_SFString(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience

	if (strcmp(r.current_word,"\"")!=0)
		return (fieldType*)r.parse_error("SFString",r.current_word);
	SFString* str;
	for (int i=0; i<BUFFERSIZE; i++) {
		r.current_word[i]=r.get_ch();
		switch(r.current_word[i]) {
		case '\\':
			r.current_word[i]=r.get_ch();
			break;
		case '"':
			r.current_word[i]='\0';
			str=new SFString(r.current_word);
			if (!str || !str->value) OUTMEM;
			return str;
			break;
		case 0:
			return (fieldType*)r.parse_error("complete SFString","EOF");
			break;
		default:
			break;
		}
	}
	return (fieldType*)r.parse_error("SFString","(string too long to read)");
}

SFString* copy_SFString(SFString* n, SFString* old)
{
	n->value=new char[strlen(old->value)+1];
	if (!n->value) OUTMEM;
	strcpy(n->value,old->value);
	return n;
}

fieldType* parse_MFString(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience

	if (strcmp(r.current_word,"\"")!=0 &&
		strcmp(r.current_word,"[")!=0)
		return (fieldType*)r.parse_error("MFString",r.current_word);

	MFString* mfstr=NULL;
	char* str;
	int multiple=FALSE, i;
	static DLList<char> buffer;

	if (r.current_word[0]=='[') {
		multiple=TRUE;
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("SFString or ']'",NULL);
	}
	for (;;) {
		if (multiple && strcmp(r.current_word,"]")==0) break;
		if (strcmp(r.current_word,"\"")!=0)
			return (fieldType*)r.parse_error("SFString",r.current_word);
		for (i=0; i<BUFFERSIZE; i++) {
			r.current_word[i]=r.get_ch();
			switch(r.current_word[i]) {
			case '\\':
				r.current_word[i]=r.get_ch();
				break;
			case '"':
				r.current_word[i]='\0';
				str=new char[strlen(r.current_word)+1];
				if (!str) OUTMEM;
				strcpy(str,r.current_word);
				if (!multiple) {
					mfstr=new MFString(1);
					if (mfstr) mfstr->value[0]=str;
					return mfstr;
				}
				buffer.add(str);
				i=BUFFERSIZE+1;
				break;
			case '\0':
				return (fieldType*)r.parse_error("complete SFString","EOF");
				break;
			default:
				break;
			}
		}
		if (i==BUFFERSIZE)
			return (fieldType*)r.parse_error("SFString","(string too long to read)");
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("SFString or ']'","EOF");
	}
	if (!mfstr) mfstr=new MFString(buffer.n_items());
	if (!mfstr) OUTMEM;
	else {
		buffer.start();
		for (i=0; i<buffer.n_items(); i++) {
			mfstr->value[i]=buffer.current();
			buffer.next();
		}
        buffer.empty();
		//    cout << "MFString has " << mfstr->n_items << " items\n";
		//    for (i=0; i<mfstr->n_items; i++)
		//	cout << "MFString contains \"" << mfstr->value[i] << "\"\n";
		return mfstr;
	}
}

MFString* copy_MFString(MFString* n, MFString* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new char*[old->n_items];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items; i++) {
			n->value[i]=new char[strlen(old->value[i])+1];
			if (!(n->value[i])) OUTMEM;
			strcpy(n->value[i],old->value[i]);
		}
	}
	return n;
}

fieldType* parse_SFTime(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFTime* obj=new SFTime();
	if (!obj) OUTMEM;
	GET_FLOAT(obj->value);
	return obj;
}

SFTime* copy_SFTime(SFTime* n, SFTime* old)
{
	n->value=old->value;
	return n;
}

fieldType* parse_MFTime(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	F_TYPE *val;
	MFTime* mft;
	static DLList<F_TYPE> buffer;
	unsigned char multiple=FALSE;
	if (strcmp(r.current_word,"[")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("time value",NULL);
		multiple=TRUE;
	}
	while (strcmp(r.current_word,"]")!=0) {
		val=new F_TYPE[1];
		if (!val) OUTMEM;
		GET_FLOAT(*val);
		buffer.add(val);
		if (multiple) {
			if (!r.get_next_word())
				return (fieldType*)r.parse_error("time value or ]",NULL);
		} else
			break;
	}
	mft=new MFTime(buffer.n_items());
	if (!mft) OUTMEM;
	buffer.start();
	for (int i=0; i<buffer.n_items(); i++) {
		mft->value[i]=*(buffer.current());
		buffer.next();
	}
    buffer.empty_and_delete_all();
	return mft;
}

MFTime* copy_MFTime(MFTime* n, MFTime* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new F_TYPE[old->n_items];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items; i++) {
			n->value[i]=old->value[i];
		}
	}
	return n;
}

fieldType* parse_SFVec2f(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFVec2f* obj=new SFVec2f();
	if (!obj) OUTMEM;
	GET_FLOAT(obj->value[0]);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("y value",NULL);
	GET_FLOAT(obj->value[1]);
	return obj;
}

SFVec2f* copy_SFVec2f(SFVec2f* n, SFVec2f* old)
{
	for (int i=0; i<2; i++) n->value[i]=old->value[i];
	return n;
}

fieldType* parse_MFVec2f(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	F_TYPE *sfv;
	MFVec2f* mfv;
	static DLList<F_TYPE> buffer;
	unsigned char multiple=FALSE;
	if (strcmp(r.current_word,"[")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("x value",NULL);
		multiple=TRUE;
	}
	while (strcmp(r.current_word,"]")!=0) {
		sfv=new F_TYPE[2];
		if (!sfv) OUTMEM;
		GET_FLOAT(sfv[0]);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("y value",NULL);
		GET_FLOAT(sfv[1]);
		buffer.add(sfv);
		if (multiple) {
			if (!r.get_next_word())
				return (fieldType*)r.parse_error("x value or ]",NULL);
		} else
			break;
	}
	mfv=new MFVec2f(buffer.n_items());
	if (!mfv) OUTMEM;
	buffer.start();
	for (int i=0; i<buffer.n_items(); i++) {
		mfv->value[2*i]=buffer.current()[0];
		mfv->value[2*i+1]=buffer.current()[1];
		buffer.next();
	}
    buffer.empty_and_delete_all();
	return mfv;
}

MFVec2f* copy_MFVec2f(MFVec2f* n, MFVec2f* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new F_TYPE[old->n_items*2];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items*2; i++) n->value[i]=old->value[i];
	}
	return n;
}

fieldType* parse_SFVec3f(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	SFVec3f* obj=new SFVec3f();
	if (!obj) OUTMEM;
	GET_FLOAT(obj->value[0]);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("y value",NULL);
	GET_FLOAT(obj->value[1]);
	if (!r.get_next_word())
		return (fieldType*)r.parse_error("z value",NULL);
	GET_FLOAT(obj->value[2]);
	return obj;
}

SFVec3f* copy_SFVec3f(SFVec3f* n, SFVec3f* old)
{
	for (int i=0; i<3; i++) n->value[i]=old->value[i];
	return n;
}

fieldType* parse_MFVec3f(ParserObject* p)
{
	ReadFileObject& r=*(p->current_file); // for convenience
	F_TYPE *sfv;
	MFVec3f* mfv;
	static DLList<F_TYPE> buffer;
	unsigned char multiple=FALSE;
	if (strcmp(r.current_word,"[")==0) {
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("x value",NULL);
		multiple=TRUE;
	}
	while (strcmp(r.current_word,"]")!=0) {
		sfv=new F_TYPE[3];
		if (!sfv) OUTMEM;
		GET_FLOAT(sfv[0]);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("y value",NULL);
		GET_FLOAT(sfv[1]);
		if (!r.get_next_word())
			return (fieldType*)r.parse_error("z value",NULL);
		GET_FLOAT(sfv[2]);
		buffer.add(sfv);
		if (multiple) {
			if (!r.get_next_word())
				return (fieldType*)r.parse_error("x value or ]",NULL);
		} else
			break;
	}
	mfv=new MFVec3f(buffer.n_items());
	if (!mfv) OUTMEM;
	buffer.start();
	for (int i=0; i<buffer.n_items(); i++) {
		mfv->value[3*i]=buffer.current()[0];
		mfv->value[3*i+1]=buffer.current()[1];
		mfv->value[3*i+2]=buffer.current()[2];
		buffer.next();
	}
	buffer.empty_and_delete_all();
	return mfv;
}

MFVec3f* copy_MFVec3f(MFVec3f* n, MFVec3f* old)
{
	int i;
	n->n_items=old->n_items;
	if (n->n_items) {
		n->value=new F_TYPE[old->n_items*3];
		if (!n->value) OUTMEM;
		for (i=0; i<n->n_items*3; i++) n->value[i]=old->value[i];
	}
	return n;
}

//
// This function creates a new object, and copies an "old" object into it
//

#define CASE(NAME,TYPE) case (NAME): \
	n=new TYPE(); \
	if (!n) { \
        cerr << "Ran out of memory\n"; \
        abort(); \
	} \
	copy_ ## TYPE ((TYPE*)n,(TYPE*)old); \
break

fieldType* field_copy(fieldType* old)
{
	fieldType* n;
	switch (old->fieldTypeName) {
		CASE(SFBOOL,SFBool);
		CASE(SFCOLOR,SFColor);
		CASE(MFCOLOR,MFColor);
		CASE(SFFLOAT,SFFloat);
		CASE(MFFLOAT,MFFloat);
		CASE(SFIMAGE,SFImage);
		CASE(SFINT32,SFInt32);
		CASE(MFINT32,MFInt32);
		CASE(SFROTATION,SFRotation);
		CASE(MFROTATION,MFRotation);
		CASE(SFSTRING,SFString);
		CASE(MFSTRING,MFString);
		CASE(SFTIME,SFTime);
		CASE(MFTIME,MFTime);
		CASE(SFVEC2F,SFVec2f);
		CASE(MFVEC2F,MFVec2f);
		CASE(SFVEC3F,SFVec3f);
		CASE(MFVEC3F,MFVec3f);
    default:
		cerr << "Internal error: field_copy called for unknown type "
			<< int(old->fieldTypeName) << '\n';
        abort();
		break;
	}
	return n;
}
