/*
* stdnodes.cpp
*
* $Id: stdnodes.cpp,v 1.4 2003/02/18 23:23:54 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*          http://www.chemicalgraphics.com/paul
*
* This file contains functions that are used to define standard nodes.
* It is heavily macro'ed so that I didn't have to type as much... The
* (node)_types functions return the expected types for fields in each
* node, used for type checking during parsing. The finish_(node) functions
* fill in a node with default values for fields not specified after the
* node - in other words, each node stored in memory will have *all* its
* fields stored, not just non-default ones. I chose to do that to make my
* life easier when creating the POV objects, but this isn't strictly
* necessary.
*   Note a big exception is events - no event (in or out) info is stored,
* since I'm using this to get a static picture. But that could be changed
* easily enough.
*   This program is of coursed biased towards nodes that contain information
* useful to a POV scene, so not all types of nodes are implemented or stored
* in the memory model. Warnings are given for unimplemented nodes.
*/

#include "fieldtyp.h"
#include "node.h"

#define __vrml2pov_stdnodes_cc_
#include "stdnodes.h"

using namespace std;

int finish_error(SFNode*)
{
	cerr << "Internal error: finish_error called\n";
	return 0;
}
int type_error(const char*)
{
	cerr << "Internal error: type_error called\n";
	return 0;
}

int (*finish_function)(SFNode*)=finish_error;
int (*type_function)(const char*)=type_error;

#define ISTYPE(TYPE,NAME) if (strcmp(id,#NAME )==0) return TYPE

#define FIND(VAR) if (strcmp(node->nodeBody.current()->fieldId,#VAR)==0) \
    found_ ## VAR =TRUE

#define DEFAULT(NAME,VAL) if (!found_ ## NAME ) { \
    node->nodeBody.add(new nodeVariable(#NAME ,new VAL)); \
    if (!(node->nodeBody.current()) || \
        !(node->nodeBody.current()->fieldValue) ) { \
        cerr << "Ran out of memory\n"; \
        abort(); \
    } \
}

// it's okay for an SFNode field to have a NULL value.

#define DEFAULTSFNODE(NAME,VAL) if (!found_ ## NAME ) { \
    node->nodeBody.add(new nodeVariable(#NAME ,VAL)); \
    if (!(node->nodeBody.current())) { \
        cerr << "Ran out of memory\n"; \
        abort(); \
    } \
}

int Appearance_types(const char* id)
{
	ISTYPE(SFNODE,material);
	ISTYPE(SFNODE,texture);
	ISTYPE(SFNODE,textureTransform);
	return ERROR;
}

int finish_Appearance(SFNode* node)
{
	int found_material=FALSE, found_texture=FALSE, found_textureTransform=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(material);
		FIND(texture);
		FIND(textureTransform);
		node->nodeBody.next();
	}
	DEFAULTSFNODE(material,NULL);
	DEFAULTSFNODE(texture,NULL);
	DEFAULTSFNODE(textureTransform,NULL);
	return 1;
}

int Box_types(const char* id)
{
	ISTYPE(SFVEC3F,size);
	return ERROR;
}

int finish_Box(SFNode* node)
{
	int found_size=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(size);
		node->nodeBody.next();
	}
	DEFAULT(size,SFVec3f(2,2,2));
	return 1;
}

int Collision_types(const char* id)
{
	ISTYPE(EVENT,addChildren);
	ISTYPE(EVENT,removeChildren);
	ISTYPE(MFNODE,children);
	ISTYPE(SFBOOL,collide);
	ISTYPE(SFVEC3F,bboxCenter);
	ISTYPE(SFVEC3F,bboxSize);
	ISTYPE(SFNODE,proxy);
	ISTYPE(EVENT,collideTime);
	return ERROR;
}

int finish_Collision(SFNode* node)
{
	int found_children=FALSE, found_collide=FALSE, found_bboxCenter=FALSE,
		found_bboxSize=FALSE, found_proxy=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(children);
		FIND(collide);
		FIND(bboxCenter);
		FIND(bboxSize);
		FIND(proxy);
		node->nodeBody.next();
	}
	DEFAULT(children,MFNode(0));
	DEFAULT(collide,SFBool(TRUE));
	DEFAULT(bboxCenter,SFVec3f(0,0,0));
	DEFAULT(bboxSize,SFVec3f(-1,-1,-1));
	DEFAULTSFNODE(proxy,NULL);
	return 1;
}

int Color_types(const char* id)
{
	ISTYPE(MFCOLOR,color);
	return ERROR;
}

int finish_Color(SFNode* node)
{
	int found_color=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(color);
		node->nodeBody.next();
	}
	DEFAULT(color,MFColor(0));
	return 1;
}

int Cone_types(const char* id)
{
	ISTYPE(SFFLOAT,bottomRadius);
	ISTYPE(SFFLOAT,height);
	ISTYPE(SFBOOL,side);
	ISTYPE(SFBOOL,bottom);
	return ERROR;
}

int finish_Cone(SFNode* node)
{
	int found_bottomRadius=FALSE, found_height=FALSE, found_side=FALSE,
		found_bottom=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(bottomRadius);
		FIND(height);
		FIND(side);
		FIND(bottom);
		node->nodeBody.next();
	}
	DEFAULT(bottomRadius,SFFloat(1));
	DEFAULT(height,SFFloat(2));
	DEFAULT(side,SFBool(TRUE));
	DEFAULT(bottom,SFBool(TRUE));
	return 1;
}

int Coordinate_types(const char* id)
{
	ISTYPE(MFVEC3F,point);
	return ERROR;
}

int finish_Coordinate(SFNode* node)
{
	int found_point=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(point);
		node->nodeBody.next();
	}
	DEFAULT(point,MFVec3f(0));
	return 1;
}

int Cylinder_types(const char* id)
{
	ISTYPE(SFBOOL,bottom);
	ISTYPE(SFFLOAT,height);
	ISTYPE(SFFLOAT,radius);
	ISTYPE(SFBOOL,side);
	ISTYPE(SFBOOL,top);
	return ERROR;
}

int finish_Cylinder(SFNode* node)
{
	int found_bottom=FALSE, found_height=FALSE, found_radius=FALSE,
		found_side=FALSE, found_top=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(bottom);
		FIND(height);
		FIND(radius);
		FIND(side);
		FIND(top);
		node->nodeBody.next();
	}
	DEFAULT(bottom,SFBool(TRUE));
	DEFAULT(height,SFFloat(2));
	DEFAULT(radius,SFFloat(1));
	DEFAULT(side,SFBool(TRUE));
	DEFAULT(top,SFBool(TRUE));
	return 1;
}

int Extrusion_types(const char* id)
{
	ISTYPE(EVENT,set_crossSection);
	ISTYPE(EVENT,set_orientation);
	ISTYPE(EVENT,set_scale);
	ISTYPE(EVENT,set_spine);
	ISTYPE(SFBOOL,beginCap);
	ISTYPE(SFBOOL,ccw);
	ISTYPE(SFBOOL,convex);
	ISTYPE(SFFLOAT,creaseAngle);
	ISTYPE(MFVEC2F,crossSection);
	ISTYPE(SFBOOL,endCap);
	ISTYPE(MFROTATION,orientation);
	ISTYPE(MFVEC2F,scale);
	ISTYPE(SFBOOL,solid);
	ISTYPE(MFVEC3F,spine);
	return ERROR;
}

int finish_Extrusion(SFNode* node)
{
	int found_beginCap=FALSE, found_ccw=FALSE, found_convex=FALSE,
		found_creaseAngle=FALSE, found_crossSection=FALSE, found_endCap=FALSE,
		found_orientation=FALSE, found_scale=FALSE, found_solid=FALSE,
		found_spine=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(beginCap);
		FIND(ccw);
		FIND(convex);
		FIND(creaseAngle);
		FIND(crossSection);
		FIND(endCap);
		FIND(orientation);
		FIND(scale);
		FIND(solid);
		FIND(spine);
		node->nodeBody.next();
	}
	DEFAULT(beginCap,SFBool(TRUE));
	DEFAULT(ccw,SFBool(TRUE));
	DEFAULT(convex,SFBool(TRUE));
	DEFAULT(creaseAngle,SFFloat(0));
	if (!found_crossSection) {
		node->nodeBody.add(new nodeVariable("crossSection",new MFVec2f(5)));
		if (!(node->nodeBody.current()) ||
			!(node->nodeBody.current()->fieldValue) ) OUTMEM;
        F_TYPE *arr=((MFVec2f*)(node->nodeBody.current()->fieldValue))->value;
        arr[0]=1; arr[1]=1; arr[2]=1; arr[3]=-1; arr[4]=-1; arr[5]=-1; 
        arr[6]=-1; arr[7]=1; arr[8]=1; arr[9]=1; 
	}
	DEFAULT(endCap,SFBool(TRUE));
	if (!found_orientation) {
		node->nodeBody.add(new nodeVariable("orientation",new MFRotation(1)));
		if (!(node->nodeBody.current()) ||
			!(node->nodeBody.current()->fieldValue) ) OUTMEM;
        F_TYPE *arr=((MFRotation*)(node->nodeBody.current()->fieldValue))->value;
        arr[0]=0; arr[1]=0; arr[2]=1; arr[3]=0; 
	}
	if (!found_scale) {
		node->nodeBody.add(new nodeVariable("scale",new MFVec2f(1)));
		if (!(node->nodeBody.current()) ||
			!(node->nodeBody.current()->fieldValue) ) OUTMEM;
        F_TYPE *arr=((MFVec2f*)(node->nodeBody.current()->fieldValue))->value;
        arr[0]=1; arr[1]=1; 
	}
	DEFAULT(solid,SFBool(TRUE));
	if (!found_spine) {
		node->nodeBody.add(new nodeVariable("spine",new MFVec3f(2)));
		if (!(node->nodeBody.current()) ||
			!(node->nodeBody.current()->fieldValue) ) OUTMEM;
        F_TYPE *arr=((MFVec3f*)(node->nodeBody.current()->fieldValue))->value;
        arr[0]=0; arr[1]=0; arr[2]=0; arr[3]=0; arr[4]=1; arr[5]=0; 
	}
	return 1;
}

int Group_types(const char* id)
{
	ISTYPE(EVENT,addChildren);
	ISTYPE(EVENT,removeChildren);
	ISTYPE(MFNODE,children);
	ISTYPE(SFVEC3F,bboxCenter);
	ISTYPE(SFVEC3F,bboxSize);
	return ERROR;
}

int finish_Group(SFNode* node)
{
	int found_children=FALSE, found_bboxCenter=FALSE, found_bboxSize=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(children);
		FIND(bboxCenter);
		FIND(bboxSize);
		node->nodeBody.next();
	}
	DEFAULT(children,MFNode(0));
	DEFAULT(bboxCenter,SFVec3f(0,0,0));
	DEFAULT(bboxSize,SFVec3f(-1,-1,-1));
	return 1;
}

int IndexedFaceSet_types(const char* id)
{
	ISTYPE(EVENT,set_colorIndex);
	ISTYPE(EVENT,set_coordIndex);
	ISTYPE(EVENT,set_normalIndex);
	ISTYPE(EVENT,set_texCoordIndex);
	ISTYPE(SFNODE,color);
	ISTYPE(SFNODE,coord);
	ISTYPE(SFNODE,normal);
	ISTYPE(SFNODE,texCoord);
	ISTYPE(SFBOOL,ccw);
	ISTYPE(MFINT32,colorIndex);
	ISTYPE(SFBOOL,colorPerVertex);
	ISTYPE(SFBOOL,convex);
	ISTYPE(MFINT32,coordIndex);
	ISTYPE(SFFLOAT,creaseAngle);
	ISTYPE(MFINT32,normalIndex);
	ISTYPE(SFBOOL,normalPerVertex);
	ISTYPE(SFBOOL,solid);
	ISTYPE(MFINT32,texCoordIndex);
	return ERROR;
}

int finish_IndexedFaceSet(SFNode* node)
{
	int found_color=FALSE, found_coord=FALSE, found_normal=FALSE,
		found_texCoord=FALSE, found_ccw=FALSE, found_colorIndex=FALSE,
		found_colorPerVertex=FALSE, found_convex=FALSE, found_coordIndex=FALSE,
		found_creaseAngle=FALSE, found_normalIndex=FALSE,
		found_normalPerVertex=FALSE, found_solid=FALSE,
		found_texCoordIndex=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(color);
		FIND(coord);
		FIND(normal);
		FIND(texCoord);
		FIND(ccw);
		FIND(colorIndex);
		FIND(colorPerVertex);
		FIND(convex);
		FIND(coordIndex);
		FIND(creaseAngle);
		FIND(normalIndex);
		FIND(normalPerVertex);
		FIND(solid);
		FIND(texCoordIndex);
		node->nodeBody.next();
	}
	DEFAULTSFNODE(color,NULL);
	DEFAULTSFNODE(coord,NULL);
	DEFAULTSFNODE(normal,NULL);
	DEFAULTSFNODE(texCoord,NULL);
	DEFAULT(ccw,SFBool(TRUE));
	DEFAULT(colorIndex,MFInt32(0));
	DEFAULT(colorPerVertex,SFBool(TRUE));
	DEFAULT(convex,SFBool(TRUE));
	DEFAULT(coordIndex,MFInt32(0));
	DEFAULT(creaseAngle,SFFloat(0));
	DEFAULT(normalIndex,MFInt32(0));
	DEFAULT(normalPerVertex,SFBool(TRUE));
	DEFAULT(solid,SFBool(TRUE));
	DEFAULT(texCoordIndex,MFInt32(0));
	return 1;
}

int IndexedLineSet_types(const char* id)
{
	ISTYPE(EVENT,set_colorIndex);
	ISTYPE(EVENT,set_coordIndex);
	ISTYPE(SFNODE,color);
	ISTYPE(SFNODE,coord);
	ISTYPE(MFINT32,colorIndex);
	ISTYPE(SFBOOL,colorPerVertex);
	ISTYPE(MFINT32,coordIndex);
	return ERROR;
}

int finish_IndexedLineSet(SFNode* node)
{
	int found_color=FALSE, found_coord=FALSE, found_colorIndex=FALSE,
		found_colorPerVertex=FALSE, found_coordIndex=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(color);
		FIND(coord);
		FIND(colorIndex);
		FIND(colorPerVertex);
		FIND(coordIndex);
		node->nodeBody.next();
	}
	DEFAULTSFNODE(color,NULL);
	DEFAULTSFNODE(coord,NULL);
	DEFAULT(colorIndex,MFInt32(0));
	DEFAULT(colorPerVertex,SFBool(TRUE));
	DEFAULT(coordIndex,MFInt32(0));
	return 1;
}

int Inline_types(const char* id)
{
	ISTYPE(MFSTRING,url);
	ISTYPE(SFVEC3F,bboxCenter);
	ISTYPE(SFVEC3F,bboxSize);
	return ERROR;
}

int finish_Inline(SFNode* node)
{
	int found_url=FALSE, found_bboxCenter=FALSE, found_bboxSize=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(url);
		FIND(bboxCenter);
		FIND(bboxSize);
		node->nodeBody.next();
	}
	DEFAULT(url,MFString(0));
	DEFAULT(bboxCenter,SFVec3f(0,0,0));
	DEFAULT(bboxSize,SFVec3f(-1,-1,-1));
	return 1;
}

int Material_types(const char* id)
{
	ISTYPE(SFFLOAT,ambientIntensity);
	ISTYPE(SFCOLOR,diffuseColor);
	ISTYPE(SFCOLOR,emissiveColor);
	ISTYPE(SFFLOAT,shininess);
	ISTYPE(SFCOLOR,specularColor);
	ISTYPE(SFFLOAT,transparency);
	return ERROR;
}

int finish_Material(SFNode* node)
{
	int found_ambientIntensity=FALSE, found_diffuseColor=FALSE,
		found_emissiveColor=FALSE, found_shininess=FALSE,
		found_specularColor=FALSE, found_transparency=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(ambientIntensity);
		FIND(diffuseColor);
		FIND(emissiveColor);
		FIND(shininess);
		FIND(specularColor);
		FIND(transparency);
		node->nodeBody.next();
	}
	DEFAULT(ambientIntensity,SFFloat(0.2));
	DEFAULT(diffuseColor,SFColor(0.8,0.8,0.8));
	DEFAULT(emissiveColor,SFColor(0,0,0));
	DEFAULT(shininess,SFFloat(0.2));
	DEFAULT(specularColor,SFColor(0,0,0));
	DEFAULT(transparency,SFFloat(0));
	return 1;
}

int Normal_types(const char* id)
{
	ISTYPE(MFVEC3F,vector);
	return ERROR;
}

int finish_Normal(SFNode* node)
{
	int found_vector=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(vector);
		node->nodeBody.next();
	}
	DEFAULT(vector,MFVec3f(0));
	return 1;
}

int PointSet_types(const char* id)
{
	ISTYPE(SFNODE,color);
	ISTYPE(SFNODE,coord);
	return ERROR;
}

int finish_PointSet(SFNode* node)
{
	int found_color=FALSE, found_coord=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(color);
		FIND(coord);
		node->nodeBody.next();
	}
	DEFAULTSFNODE(color,NULL);
	DEFAULTSFNODE(coord,NULL);
	return 1;
}

int Shape_types(const char* id)
{
	ISTYPE(SFNODE,appearance);
	ISTYPE(SFNODE,geometry);
	return ERROR;
}

int finish_Shape(SFNode* node)
{
	int found_appearance=FALSE, found_geometry=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(appearance);
		FIND(geometry);
		node->nodeBody.next();
	}
	DEFAULTSFNODE(appearance,NULL);
	DEFAULTSFNODE(geometry,NULL);
	return 1;
}

int Sphere_types(const char* id)
{
	ISTYPE(SFFLOAT,radius);
	return ERROR;
}

int finish_Sphere(SFNode* node)
{
	int found_radius=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(radius);
		node->nodeBody.next();
	}
	DEFAULT(radius,SFFloat(1));
	return 1;
}

int Transform_types(const char* id)
{
	ISTYPE(EVENT,addChildren);
	ISTYPE(EVENT,removeChildren);
	ISTYPE(SFVEC3F,center);
	ISTYPE(MFNODE,children);
	ISTYPE(SFROTATION,rotation);
	ISTYPE(SFVEC3F,scale);
	ISTYPE(SFROTATION,scaleOrientation);
	ISTYPE(SFVEC3F,translation);
	ISTYPE(SFVEC3F,bboxCenter);
	ISTYPE(SFVEC3F,bboxSize);
	return ERROR;
}

int finish_Transform(SFNode* node)
{
	int found_center=FALSE, found_children=FALSE, found_rotation=FALSE,
		found_scale=FALSE, found_scaleOrientation=FALSE,
		found_translation=FALSE, found_bboxCenter=FALSE, found_bboxSize=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(center);
		FIND(children);
		FIND(rotation);
		FIND(scale);
		FIND(scaleOrientation);
		FIND(translation);
		FIND(bboxCenter);
		FIND(bboxSize);
		node->nodeBody.next();
	}
	DEFAULT(center,SFVec3f(0,0,0));
	DEFAULT(children,MFNode(0));
	DEFAULT(rotation,SFRotation(0,0,1,0));
	DEFAULT(scale,SFVec3f(1,1,1));
	DEFAULT(scaleOrientation,SFRotation(0,0,1,0));
	DEFAULT(translation,SFVec3f(0,0,0));
	DEFAULT(bboxCenter,SFVec3f(0,0,0));
	DEFAULT(bboxSize,SFVec3f(-1,-1,-1));
	return 1;
}

int WorldInfo_types(const char* id)
{
	ISTYPE(MFSTRING,info);
	ISTYPE(SFSTRING,title);
	return ERROR;
}

int Viewpoint_types(const char* id)
{
	ISTYPE(EVENT,set_bind);
	ISTYPE(SFFLOAT,fieldOfView);
	ISTYPE(SFBOOL,jump);
	ISTYPE(SFROTATION,orientation);
	ISTYPE(SFVEC3F,position);
	ISTYPE(SFSTRING,description);
	ISTYPE(EVENT,bindTime);
	ISTYPE(EVENT,isBound);
	return ERROR;
}

int finish_Viewpoint(SFNode* node)
{
	int found_fieldOfView=FALSE, found_jump=FALSE, found_orientation=FALSE,
		found_position=FALSE, found_description=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(fieldOfView);
		FIND(jump);
		FIND(orientation);
		FIND(position);
		FIND(description);
		node->nodeBody.next();
	}
	DEFAULT(fieldOfView,SFFloat(0.785398));
	DEFAULT(jump,SFBool(TRUE));
	DEFAULT(orientation,SFRotation(0,0,1,0));
	DEFAULT(position,SFVec3f(0,0,10));
	DEFAULT(description,SFString(""));
	return 1;
}

int finish_WorldInfo(SFNode* node)
{
	int found_info=FALSE, found_title=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
		FIND(info);
		FIND(title);
		node->nodeBody.next();
	}
	DEFAULT(info,MFString(0));
	DEFAULT(title,SFString(""));
	return 1;
}

#define FUNC(NODE) if (strcmp(id,#NODE )==0) { \
	finish_function=finish_ ## NODE ; \
	type_function= NODE ## _types ; \
	return IMPLEMENTED; \
}

#define NOTIMP(NODE) if (strcmp(id,#NODE )==0) return UNIMPLEMENTED

int find_standard_node(const char* id)
{
	finish_function=finish_error;
	type_function=type_error;
	NOTIMP(Anchor);
	FUNC(Appearance);
	NOTIMP(AudioClip);
	NOTIMP(Background);
	NOTIMP(Billboard);
	FUNC(Box);
	FUNC(Collision);
	FUNC(Color);
	NOTIMP(ColorInterpolator);
	FUNC(Cone);
	FUNC(Coordinate);
	NOTIMP(CoordinateInterpolator);
	FUNC(Cylinder);
	NOTIMP(CylinderSensor);
	NOTIMP(DirectionalLight);
	NOTIMP(ElevationGrid);
	FUNC(Extrusion);
	NOTIMP(Fog);
	NOTIMP(FontStyle);
	FUNC(Group);
	NOTIMP(ImageTexture);
	FUNC(IndexedFaceSet);
	FUNC(IndexedLineSet);
	FUNC(Inline);
	NOTIMP(LOD);
	FUNC(Material);
	NOTIMP(MovieTexture);
	NOTIMP(NavigationInfo);
	FUNC(Normal);
	NOTIMP(NormalInterpolator);
	NOTIMP(OrientationInterpolator);
	NOTIMP(PixelTexture);
	NOTIMP(PlaneSensor);
	NOTIMP(PointLight);
	FUNC(PointSet);
	NOTIMP(PositionInterpolator);
	NOTIMP(ProximitySensor);
	NOTIMP(ScalarInterpolator);
	NOTIMP(Script); // leave as NOTIMP, even though we know how to deal with it
	FUNC(Shape);
	NOTIMP(Sound);
	FUNC(Sphere);
	NOTIMP(SphereSensor);
	NOTIMP(SpotLight);
	NOTIMP(Switch);
	NOTIMP(Text);
	NOTIMP(TextureCoordinate);
	NOTIMP(TextureTransform);
	NOTIMP(TimeSensor);
	NOTIMP(TouchSensor);
	FUNC(Transform);
	FUNC(Viewpoint);
	NOTIMP(VisibilitySensor);
	FUNC(WorldInfo);
	return 0;
}

/* Fill in for new node type

  int _types(const char* id)
  {
  ISTYPE(EVENT,);
  ISTYPE(,);
  return ERROR;
  }
  
	int finish_(SFNode* node)
	{
	int found_=FALSE, found_=FALSE, found_=FALSE;
	for (int i=0; i<node->nodeBody.n_items(); i++) {
    FIND();
    FIND();
    node->nodeBody.next();
	}
	DEFAULT(,);
	DEFAULTSFNODE(,NULL);
	return 1;
	}
	
*/
