/*
* vrml2pov.cpp
*
* $Id: vrml2pov.cpp,v 1.5 2003/02/18 23:23:54 Paul Exp $
*
* Copyright (C) 1998-2003 by Paul Thiessen, Chemical Graphics
*                 paul@chemicalgraphics.com
*            http://www.chemicalgraphics.com/paul
*
* The main program - just links the parser with the POV generator.
*/

#include <iostream>

#include "parser.h"
#include "pov.h"

using namespace std;

const char* usage="\n\
Usage: vrml2pov [-v] [-t] [-h #] [-l #] [-p #] [-b] vrml_file pov_file\n\
\n\
  -v gives a verbose output (mainly for debugging)\n\
  -t prints out the final node tree after parsing and expansion of PROTO's\n\
  -h sets headlight brightness (generally from 0-1, e.g. -h 0.8;\n\
     if not specified, default is based on # other lights present)\n\
  -l sets line radius for IndexedLineSet (default 0.05)\n\
  -p sets point radius for PointSet (default 0.05)\n\
  -b turns on backface culling - requires first node to be Viewpoint\n\
     (and currently only applies to IndexedFaceSet)\n";

void print_out_node(SFNode*);

int main(int argc, char *argv[])
{
	ParserObject parser;	// the VRML scene is stored in here
	int i, print_node_tree=FALSE;

	// process command line options

	for (i=1; i<argc; i++) {
		if (argv[i][0]!='-') break;
		switch (argv[i][1]) {
		case 'v':
			parser.verbose();
			cout << VERSION;
			break;
		case 't':
			print_node_tree=TRUE;
			break;
		case 'h':
			if ((i+1)>=argc) {
				cerr << usage;
				return 1;
			}
			headlight=atof(argv[i+1]);
			i++;
			break;
		case 'l':
			if ((i+1)>=argc) {
				cerr << usage;
				return 1;
			}
			line_radius=atof(argv[i+1]);
			i++;
			break;
		case 'p':
			if ((i+1)>=argc) {
				cerr << usage;
				return 1;
			}
			point_radius=atof(argv[i+1]);
			i++;
			break;
		case 'b':
			do_backface_culling=TRUE;
			break;
		default:
			cerr << "Unknown option " << argv[i] << '\n';
			cerr << usage;
			return 1;
		}
	}

	//  parse the input file

	if (argc<(i+2)) {
		cerr << VERSION << usage;
		return 1;
	}
	if (parser.parse(argv[i])) {
		cerr << "Parsing error. Aborting.\n";
		return 1;
	}

	// output the root node tree (i.e. the whole scene as parsed)

	if (print_node_tree) {
        cout << "#VRML V2.0 utf8\n";
		parser.root_Nodes.start();
		for (int j=0; j<(parser.root_Nodes.n_items()); j++) {
			print_out_node(parser.root_Nodes.current());
			parser.root_Nodes.next();
		}
	}

	// create the POV file (obviously...)

	do_POV_output(&parser,argv[i],argv[i+1]);

	return 0;
}

#define SPACE cout.width(indent); cout << ""

void print_out_node(SFNode* node)
{
	static int indent=0;
	static const int ii=2;  // indendation increment
    int i, j, l;
    ios::fmtflags opt;

	SPACE;
	cout << node->nodeTypeId << " {";
	if (node->nodeBody.n_items()==0)
		cout << " }\n";
	else {
		cout << '\n';
		indent+=ii;
		node->nodeBody.start();
		for (i=0; i<(node->nodeBody.n_items()); i++) {
			SPACE;
			cout << node->nodeBody.current()->fieldId << ' ';
			fieldType* field=node->nodeBody.current()->fieldValue;
			if (!field)
				cout << "NULL\n";
			else switch (field->fieldTypeName) {

      case SFBOOL:
		  if (((SFBool*)field)->value)
			  cout << "TRUE\n";
		  else
			  cout << "FALSE\n";
		  break;

      case SFCOLOR:
		  cout.precision(4);
		  cout << ((SFColor*)field)->value[0] << ' '
			  << ((SFColor*)field)->value[1] << ' '
			  << ((SFColor*)field)->value[2] << '\n';
		  break;

      case MFCOLOR:
		  cout.precision(4);
		  if ((((MFColor*)field)->n_items)==0)
			  cout << "[]\n";
		  else {
              cout << "[\n";
              SPACE;
			  for (j=0; j<(((MFColor*)field)->n_items); j++) {
				  cout << ((MFColor*)field)->value[3*j] << ' '
					  << ((MFColor*)field)->value[3*j+1] << ' '
                      << ((MFColor*)field)->value[3*j+2] << ",\n";
                  SPACE;
			  }
			  cout << "]\n";
		  }
		  break;

      case SFFLOAT:
		  cout.precision(6);
		  cout << ((SFFloat*)field)->value << '\n';
		  break;

      case MFFLOAT:
		  cout.precision(6);
          cout << "[\n";
          SPACE;
		  for (j=0; j<(((MFFloat*)field)->n_items); j++) {
			  cout << ((MFFloat*)field)->value[j] << ' ';
              if ((j+1)%10 == 0) {
                  cout << '\n';
                  SPACE;
              }
		  }
		  cout << "]\n";
		  break;

      case SFIMAGE:
		  cout << ((SFImage*)field)->width << ' '
			  << ((SFImage*)field)->height << ' '
			  << ((SFImage*)field)->n_components;
		  l=(((SFImage*)field)->width)*(((SFImage*)field)->height);
		  opt=cout.flags();
		  cout.setf(ios::hex,ios::basefield);
		  cout.setf(ios::showbase);
		  for (j=0; j<l; j++) {
			  cout << ' ' << ((SFImage*)field)->pixel_data[j];
		  }
		  cout.flags(opt);
		  cout << '\n';
		  break;

      case SFINT32:
		  cout << ((SFInt32*)field)->value << '\n';
		  break;

      case MFINT32:
          cout << "[\n";
          SPACE;
		  for (j=0; j<(((MFInt32*)field)->n_items); j++) {
			  cout << ((MFInt32*)field)->value[j] << ' ';
              if ((j+1)%10 == 0) {
                  cout << '\n';
                  SPACE;
              }
		  }
		  cout << "]\n";
		  break;

      case SFNODE:
		  cout << '\n';
		  indent+=ii;
		  print_out_node((SFNode*)field);
		  indent-=ii;
		  break;

      case MFNODE:
		  if ((((MFNode*)field)->n_items)==0)
			  cout << "[ ]\n";
		  else {
			  cout << "[\n";
			  indent+=ii;
			  for (j=0; j<(((MFNode*)field)->n_items); j++) {
				  print_out_node(((MFNode*)field)->value[j]);
			  }
			  indent-=ii;
			  SPACE;
			  cout << "]\n";
		  }
		  break;

      case SFROTATION:
		  cout.precision(6);
		  cout << ((SFRotation*)field)->value[0] << ' '
			  << ((SFRotation*)field)->value[1] << ' '
			  << ((SFRotation*)field)->value[2] << ' '
			  << ((SFRotation*)field)->value[3] << '\n';
		  break;

      case MFROTATION:
		  cout.precision(6);
		  if ((((MFRotation*)field)->n_items)==0)
			  cout << "[]\n";
		  else {
              cout << "[\n";
              SPACE;
			  for (j=0; j<(((MFRotation*)field)->n_items); j++) {
				  cout << ((MFRotation*)field)->value[4*j] << ' '
					  << ((MFRotation*)field)->value[4*j+1] << ' '
					  << ((MFRotation*)field)->value[4*j+2] << ' '
                      << ((MFRotation*)field)->value[4*j+3] << ",\n";
                  SPACE;
			  }
			  cout << "]\n";
		  }
		  break;

      case SFSTRING:
		  cout << '"' << ((SFString*)field)->value << "\"\n";
		  break;

      case MFSTRING:
		  if ((((MFString*)field)->n_items)==0)
			  cout << "[]\n";
		  else {
			  cout << "[\n";
			  indent+=ii;
			  for (j=0; j<(((MFString*)field)->n_items); j++) {
				  SPACE;
				  cout << '"' << ((MFString*)field)->value[j] << "\"\n";
			  }
			  indent-=ii;
			  SPACE;
			  cout << "]\n";
		  }
		  break;

      case SFTIME:
		  cout.precision(6);
		  cout << ((SFTime*)field)->value << '\n';
		  break;

      case MFTIME:
		  cout.precision(6);
          cout << "[\n";
          SPACE;
		  for (j=0; j<(((MFTime*)field)->n_items); j++) {
			  cout << ((MFTime*)field)->value[j] << ' ';
              if ((j+1)%10 == 0) {
                  cout << '\n';
                  SPACE;
              }
		  }
		  cout << "]\n";
		  break;

      case SFVEC2F:
		  cout.precision(6);
		  cout << ((SFVec2f*)field)->value[0] << ' '
			  << ((SFVec2f*)field)->value[1] << '\n';
		  break;

      case MFVEC2F:
		  cout.precision(6);
		  if ((((MFVec2f*)field)->n_items)==0)
			  cout << "[]\n";
		  else {
              cout << "[\n";
              SPACE;
			  for (j=0; j<(((MFVec2f*)field)->n_items); j++) {
				  cout << ((MFVec2f*)field)->value[2*j] << ' '
                      << ((MFVec2f*)field)->value[2*j+1] << ",\n";
                  SPACE;
			  }
			  cout << "]\n";
		  }
		  break;

      case SFVEC3F:
		  cout.precision(6);
		  cout << ((SFVec3f*)field)->value[0] << ' '
			  << ((SFVec3f*)field)->value[1] << ' '
			  << ((SFVec3f*)field)->value[2] << '\n';
		  break;

      case MFVEC3F:
		  cout.precision(6);
		  if ((((MFVec3f*)field)->n_items)==0)
			  cout << "[]\n";
		  else {
              cout << "[\n";
              SPACE;
			  for (j=0; j<(((MFVec3f*)field)->n_items); j++) {
				  cout << ((MFVec3f*)field)->value[3*j] << ' '
					  << ((MFVec3f*)field)->value[3*j+1] << ' '
                      << ((MFVec3f*)field)->value[3*j+2] << ",\n";
                  SPACE;
			  }
			  cout << "]\n";
		  }
		  break;

      default:
		  cout << "(some value)\n";
		  break;
      }
      node->nodeBody.next();
    }
    indent-=ii;
    SPACE;
    cout << "}\n";
  }
  return;
}
